class AuthenticatedController < ApplicationController
  include AuthenticationSystem
  include ErrorRenderer

    before_filter :basic_auth_challenge,
                  :set_expiry_time,
                  :record_activity_info,
                  :expire_session,
                  :check_login,
                  :check_password_expiry,
                  :check_permissions

protected

  def expire_session
    if session_timed_out?
      url = after_logout_url
      reset_session
      flash[:warning] = t('app.controllers.authenticated_controller.timed_out')
      save_intended_destination
      redirect_to url
    end
  end

  # for resetting the session timeout timer after a request
  def set_expiry_time
    unless api_request? || request.xhr?
      unless session_timed_out?
        expiry_time = EConfiguration.session_timeout.minutes.from_now
        session[:expiry_time] = expiry_time
        current_user.update_attribute(:session_expiry_time, expiry_time) if current_user
      end
    end
  end

  # for recording params from an authenticated request
  def record_activity_info
    current_user.record_activity if current_user
  end

  # make sure that this session is still the valid one for this user
  def check_login
    unless api_request? || request.xhr?
      # if this session is logged in, make sure the session UUID matches the currently stored UUID for the user
      if session[:login_user_id] && session[:login_uuid]
        _user = User.find(session[:login_user_id])
        if _user.login_uuid != session[:login_uuid]
          # Session UUID doesn't match - end session, but don't reset login UUID
          # Resetting login UUID would invalidate valid sessions too
          reset_session(false)
          if EConfiguration.prevent_concurrent_login?
            flash[:warning] = t('app.controllers.authenticated_controller.new_session', {username: (ERB::Util.h _user.name), remote: (ERB::Util.h _user.last_remote_addr)})
          else
            flash[:warning] = t('app.controllers.authenticated_controller.expired')
          end
          redirect_to new_login_url(local: true)
        end
      end
    end
  end

  def basic_auth_challenge
    return unless api_request?
    # Challenge service requests that don't have HTTP-Basic auth headers
    if request.headers['HTTP_AUTHORIZATION'].nil?
      headers['WWW-Authenticate'] = "Basic 'REST API'"
      render_error(:unauthorized)
      return
    end

    # Return 401 if authentication failed
    render_error(:unauthorized) if current_user.nil?
  end

  def check_permissions
    logged_in_user = current_user

    #if bad cookie
    render_response(450) and return if logged_in_user.nil? && params[:format]=='xml'

    if logged_in_user.nil?
      check_deactivated_users
      save_intended_destination
      handle_not_logged_in
      return #terminate rest of filter chain
    end
  end

  # Will set params[:local] to true if the session_user was deactivated during use, such that the user will be redirected to local login
  def check_deactivated_users
    session_user = AuthorizableUser.inactive.find_by_id(session[:login_user_id])
    params[:local] = true if session_user && ['LocalUser', 'SelfRegisteredUser'].include?(session_user.type)
  end

  def handle_not_logged_in
    respond_to do |format|
      format.html { redirect_to new_login_url(local: params[:local]) }
      format.xml { render_response(401) }
      format.json { render_error(:unauthorized) }
    end
  end

private

  def session_timed_out?
    session[:expiry_time].present? && session[:expiry_time] < Time.now
  end
end
