class Delivery < ActiveRecord::Base
  include MetadataContainer
  include UrlGeneration
  include ActionController::UrlWriter # for generating paths in notification emails
  extend ActiveSupport::Memoizable
  include Packages::TransferInfo
  include Packages::FileBrowsing

  attr_accessible :title, :note, :sources, :sender,
      :recipients, :private_recipients,
      :notifiable_on_upload, :notifiable_on_download, :notify_on_upload, :notify_on_download,
                  :notifiable_on_receipt, :notify_on_receipt,
      :metadata, :delete_after_download_policy,
      :use_encryption_at_rest, :skip_metadata_validation, :skip_path_validation,
      :e_package, :disallow_dropbox, :forward, :ear_passphrase,
      :uuid, :e_package_id, :sender_id, :sender_name, :invitation,
      :recipient_string, :relay, :metadata_json, :skip_source_validation,
      :package_uuid, :error_code, :error_message, :skip_recipient_validation,
      :transfer_type, :warnings, :ear_enabled, :tos_for_sending, :obfuscation_enabled, :use_obfuscation, :prevent_http_download

  # legacy API option that we just ignore
  attr_accessible :send_upload_result

  belongs_to :e_package
  belongs_to :delivery_completion, foreign_key: :sequence_id

  has_many :user_deliveries
  has_many :workgroup_deliveries
  has_many :recipient_deliveries, :dependent => :destroy
  has_many :notifications, dependent: :destroy
  has_many :transfer_requests, :dependent => :destroy
  has_many :e_sessions
  has_many :upload_sessions,
    through: :e_package,
    class_name: 'ESession',
    conditions: ["direction = ? OR direction = ?", ESession::DIRECTION_UPLOAD, ESession::DIRECTION_CREATE]
  has_many :upload_requests,
    through: :e_package,
    class_name: 'TransferRequest',
    conditions: ["direction = ? OR direction = ?", ESession::DIRECTION_UPLOAD, ESession::DIRECTION_CREATE]

  has_many :package_relays, :conditions => ["relay_type = '0'"]
  has_one  :invitation

  belongs_to :sender, :class_name => 'User', :foreign_key => 'sender_id'
  named_scope :completed, :joins => :e_package, :conditions => ["e_packages.e_package_upload_status = ?", "Completed"],
    :order => "GREATEST(e_packages.e_package_upload_completion_time, deliveries.created_at) ASC"

  before_validation_on_create :make_package, :make_recipient_deliveries, :process_metadata
  before_save :generate_sequence_id, if: :forward?
  before_save :set_sender_name
  validate_on_create :check_sender_allowed_to_create
  validate :check_presence_of_recipients, :unless => Proc.new { |a| a.skip_recipient_validation }
  validate :check_dropbox_only_recipient
  validate :check_recipients, :unless => Proc.new { |a| a.skip_recipient_validation }
  validate :check_notifications
  validate :check_sources, :unless => Proc.new { |a| a.skip_source_validation }
  validate :check_metadata, :unless => Proc.new { |a| a.skip_metadata_validation }
  validate :check_paths, :unless => Proc.new { |a| a.skip_path_validation }
  validate :check_ear
  validate :sender_share_id_same, if: :forward
  validates_length_of :title, :maximum => 245
  validates_length_of :note, :maximum => 8000, :allow_nil => true
  validates_presence_of :title
  validate :check_for_quota

  validates_acceptance_of :tos_for_sending, :message => "must be checked", :if => :use_tos?
  # Hard check after all expansion is done to make sure we have at least one recipient
  validate_on_create :check_presence_of_recipients

  attr_accessor :use_encryption_at_rest, :sources, :metadata,
      :delete_after_download_policy, :skip_metadata_validation,
      :skip_path_validation, :disallow_dropbox,
      :invalid_recipient_names, :invalid_notifications, :ear_passphrase, :skip_source_validation,
      :package_uuid, :notify_on_upload, :notify_on_download, :notify_on_receipt, :error_code, :error_message, :skip_recipient_validation,
      :transfer_type, :warnings, :use_obfuscation

  # legacy API option that we just ignore
  attr_accessor :send_upload_result

  validates_length_of :ear_passphrase, :minimum => 1, if: Proc.new { |d| !d.ear_passphrase.nil? && d.ear_passphrase.blank? }

  default_value_for(:uuid) {SecureRandom.uuid}
  default_value_for :use_encryption_at_rest, false
  default_value_for :skip_path_validation, false
  default_value_for :skip_source_validation, false
  default_value_for :sources, []

  default_value_for :notify_on_upload, false
  default_value_for :notify_on_download, false
  default_value_for :notify_on_receipt, false
  default_value_for :invalid_notifications, []

  default_value_for :invalid_recipient_names, []
  default_value_for :metadata, {}
  default_value_for :note, ""

  default_value_for :skip_recipient_validation, false
  default_value_for :warnings, {}
  default_value_for :use_obfuscation, false

  after_save :save_metadata_to_pkg_dir, if: :save_metadata?

  def after_initialize
    self.dbox_extusr_sender_pkg_passcode ||= SecureRandom.uuid if sender_id.present?
  end

  def set_sender_name
    write_attribute(:sender_name, sender_name)
  end

  def sender_name
    sender.present? ? sender.name : "(deleted user)"
  end

  def private_recipients=(str_or_arr)
    recipient_names = str_or_arr.is_a?(Array) ? str_or_arr : Utils.parse_recipients(str_or_arr)
    recipient_names = expand_distribution_lists(recipient_names).uniq
    recipient_names = expand_email_recipients(recipient_names, sender)
    recipient_names.each do |name|
      add_recipient(name, false)
    end
    str_or_arr
  end

  def public_recipients=(str_or_arr)
    recipient_names = str_or_arr.is_a?(Array) ? str_or_arr : Utils.parse_recipients(str_or_arr)
    recipient_names = expand_distribution_lists(recipient_names).uniq
    recipient_names = expand_email_recipients(recipient_names, sender)
    recipient_names.each do |name|
      add_recipient(name, true)
    end
    str_or_arr
  end

  def notifiable_on_upload=(str_or_arr)
    recipient_names = str_or_arr.is_a?(Array) ? str_or_arr : Utils.parse_recipients(str_or_arr)
    recipient_names = expand_distribution_lists(recipient_names).uniq

    write_attribute(:notifiable_on_upload, recipient_names.join(', '))
  end

  def notifiable_on_download=(str_or_arr)
    recipient_names = str_or_arr.is_a?(Array) ? str_or_arr : Utils.parse_recipients(str_or_arr)
    recipient_names = expand_distribution_lists(recipient_names).uniq

    write_attribute(:notifiable_on_download, recipient_names.join(', '))
  end

  def notifiable_on_receipt=(str_or_arr)
    recipient_names = str_or_arr.is_a?(Array) ? str_or_arr : Utils.parse_recipients(str_or_arr)
    recipient_names = expand_distribution_lists(recipient_names).uniq

    write_attribute(:notifiable_on_receipt, recipient_names.join(', '))
  end

  alias_method :recipients=, :public_recipients=


  def recipients_for_display(recipients)
    lookup_recipients(recipients).map{ |r| r.display_name rescue r }.join(', ') rescue recipients.join(', ')
  end

  def public_recipients_for_display
    recipients_for_display(public_recipients)
  end

  def private_recipients_for_display
    recipients_for_display(private_recipients)
  end

  def sender_display_name
    sender.display_name rescue sender_name
  end

  def add_recipient(name, public)
    self.invalid_recipient_names ||= []
    self.warnings ||= {}
    @recipient_hashes ||= {}
    recipient = lookup_recipient(name)
    if recipient
      @recipient_hashes[recipient] ||= {}
      @recipient_hashes[recipient][:name] = name
      @recipient_hashes[recipient][:public] ||= public
    else
      invalid_recipient_names << name
    end
  end

  def all_recipients
    recipient_deliveries.map(&:recipient_name)
  end

  def public_recipients
    recipient_deliveries.public.map(&:recipient_name)
  end

  def public_recipients_email
    lookup_recipient_emails(public_recipients)
  end

  def cc_on_receipt_recipients
    notifiable_on_receipt.present? ? notifiable_on_receipt.split(', ').uniq : []
  end

  def all_cc_recipients
    rs = []
    rs += notifiable_on_download.split(', ') if notifiable_on_download.present?
    rs += notifiable_on_upload.split(', ') if notifiable_on_upload.present?
    rs += notifiable_on_receipt.split(', ') if notifiable_on_receipt.present?

    rs.uniq
  end

  def lookup_recipient_emails(recipients)
    recipients.map{ |r|
      u = lookup_recipient(r)
      u.present? && u.is_a?(User) ? u.email : r
    }
  end

  # Used for converting faspex users to email addresses
  def notifiable_on_receipt_notifications
    notifications = []
    Utils.parse_recipients(notifiable_on_receipt).each do |name|
        notifications << build_notification(name, 'receipt')
    end

    notifications
  end

  alias_method :recipients, :public_recipients

  def private_recipients
    recipient_deliveries.private.map(&:recipient_name)
  end

  def http_transfer?
    transfer_type == 'http'
  end

  def source_share
    @source_share ||= SourceShare.find_by_id(sources.first[:id]) || e_package.source_share
  end

  def self.delete_expired_deliveries
    deliveries_to_delete = Delivery.find_by_sql("SELECT * FROM deliveries WHERE state = '#{Delivery::STATE_EXPIRED}'")

    deliveries_to_delete.each do |delivery|
      delivery.do_full_delete
    end
  end

private

  # If email is entered, send to all existing users, if any, with that email address instead
  def expand_email_recipients(recipient_names, sender)
    recipients = []

    recipient_names.each do |name|
      # Not existing user or workgroup
      if !AuthorizableUser.find_by_name(name) && !EWorkgroup.find_by_e_wg_name(name.gsub('*', ''))
        # Ignore (external) flag for now - handled by user lookup
        # Find names of all users by email that the sender can send to
        users_with_email = AuthorizableUser.find_all_by_email_and_account_activated(name, true).map{ |r| r.name }
        recipients = users_with_email.any? ? recipients + users_with_email : recipients << name
      else
        # Existing user or workgroup, don't modify
        recipients << name
      end
    end

    recipients
  end

  def expand_distribution_lists(recipient_names)
    expanded_list = []
    recipient_names.each do |name|
      # Do not create external user
      recipient = lookup_recipient(name, false)
      if recipient.is_a? DistributionList
        expanded_list += expand_distribution_lists(recipient.contact_names)
      else
        expanded_list << name
      end
    end
    expanded_list
  end

  def make_package
    return if e_package

    user_share = initial_destination

    if nil != self.metadata and metadata['SenderShareId'].present?
      id = self.metadata['SenderShareId'].to_i
      user_share = Share.find_by_id(id) || default_destination
    end

    # Make sure sources is present when making package
    unless self.sources.present?
      self.sources = [user_share]
    end

    self.e_package = EPackage.new(
        e_package_name:               title.to_s,
        e_package_sender_id:          sender.id,
        e_package_sender_name:        sender.name,
        e_package_note:               note.to_s,
        ear_encrypted:                ear_encrypted?,
        delete_after_download_policy: delete_after_download_policy,
        source_share_id:              self.sources.first && self.sources.first[:id],
        source_paths:                 self.sources.first && self.sources.first[:paths],
        storage_share:                user_share #Packages::Routing.assigned_share_for_user(self.user)
    )
    e_package.e_package_uuid = package_uuid if package_uuid
  end

  def ear_encrypted?
    # Do not allow EAR when cloud referencing
    (ear_enabled || require_ear? || ear_passphrase.present? || use_encryption_at_rest == '1') && !make_cloud_references?
  end

  def process_metadata
    if !self.metadata.empty?
      special_metadata_fields.each do |key|
        next if self.metadata[key].nil? || !self.metadata[key].is_a?(HashWithIndifferentAccess) || self.metadata[key].empty?
        self.metadata[key]=self.metadata[key].to_json
      end
    end
    self.metadata['_pkg_uuid'] = e_package.e_package_uuid
    self.metadata['_pkg_name'] = title
    self.metadata['_created_utc'] = Time.now.utc
    if dropbox_delivery?
      self.metadata['_dropbox_name'] = workgroup_deliveries.first.recipient_name
    end
    if metadata_profile
      self.metadata['_metadata_profile_name'] = metadata_profile.name
      self.metadata['_metadata_profile_uuid'] = metadata_profile.uuid
      self.metadata['_metadata_profile_version'] = metadata_profile.updated_at.utc.strftime('%Y-%m-%d %T')
    end
    self.metadata_json = metadata.to_json
  end

  def make_recipient_deliveries
    @recipient_hashes ||= {}
    @recipient_hashes.each_pair do |recipient, recipient_attributes|
      case recipient
      when EWorkgroup
        make_workgroup_delivery(recipient, recipient_attributes)
      when User
        make_user_delivery(recipient, recipient_attributes)
      else
        invalid_recipient_names << recipient_attributes[:name]
      end
    end
  end

  def make_workgroup_delivery(recipient, recipient_attributes)
    if recipient.dropbox? && disallow_dropbox
      invalid_recipient_names << recipient_attributes[:name]
    else
      workgroup_deliveries << WorkgroupDelivery.new(
        recipient:      recipient,
        recipient_name: recipient_attributes[:name],
        delivery:       self,
        e_package:      e_package,
        share:          recipient.destination_share || Share.default_destination,
        private:        !recipient_attributes[:public]
      )
    end
  end

  def make_user_delivery(recipient, recipient_attributes)

    user_share = nil

    if self.metadata and self.metadata['RecipientShareIds'] && MetadataProfile.validate_v4_field('RecipientShareIds', self.metadata['RecipientShareIds']).empty?
      user_id_to_share_id_hash = self.parse_json_field(metadata['RecipientShareIds'])
      if user_id_to_share_id_hash
        # do not perform a relay for those UserIds who are not actual recipients (i.e in the "To" field when creating the package)
        # check for case insensitive match of recipient name
        user_id_to_share_id_hash.each do |name, share|
          if name.casecmp(recipient.name) == 0
            user_share = Share.find_by_id share
            break
          end
        end
        if nil == user_share
          logger.warn("Package routing: RecipientShareIds field of metadata does not contain share ID, using default.")
        end
      end
    end
 
    user_share = storage_share if not user_share
    user_deliveries << UserDelivery.new(
      recipient:      recipient,
      recipient_name: recipient_attributes[:name],
      delivery:       self,
      e_package:      e_package,
      share:          user_share,
      private:        !recipient_attributes[:public]
    )
  end

  def lookup_recipient(name, include_external=true, for_ccs=false)
    name = name.sub(/^\*/, '')
    sender.lookup_recipient(name, include_external, for_ccs) rescue nil
  end

  # Check sender quota limit for the user before initiating delivery
  def check_for_quota
    if sender && sender.user_quota_enabled? && !sender.recipient_email_whitelisted?(self.recipient_emails.split(',')) && !sender.recipient_email_whitelisted?(self.wg_recipient_emails)
      if sender && sender.completed_session_data >= sender.user_quota_limit
        self.errors.add(:base, I18n.t('app.models.delivery.send_fail_quota_limit', limit: sender.user_quota_limit))
      else
        self.e_package.quota_contribution = true
        self.e_package.save
      end
    end
  end

  def check_sender_allowed_to_create
    unless !sender || sender.allowed_to_send_normal_packages? || dropbox.present? || (forward? && sender.forwarding_allowed)
      errors.add(:base, I18n.t('app.models.delivery.send_fail_permissions'))
    end
  end

  def check_presence_of_recipients
    @recipient_hashes ||= {}
    if @recipient_hashes.empty?
      errors.add(:base, I18n.t('app.models.delivery.send_fail_invalid_recipients')) unless errors[:base].present? && errors[:base].include?( I18n.t('app.models.delivery.send_fail_invalid_recipients'))
    end
  end

  def check_dropbox_only_recipient
    @recipient_hashes ||= {}

    dropbox_recipient = workgroup_deliveries.any?{ |wd| wd.recipient.is_a?(EWorkgroup) && wd.recipient.dropbox? }

    if dropbox_recipient.present? && @recipient_hashes.length > 1
      errors.add(:base, I18n.t('app.models.delivery.send_fail_dropbox_only'))
    end
  end

  def check_recipients
    self.invalid_recipient_names ||= []
    self.warnings ||= {}
    invalid_recipient_names.each do |name|
      warnings[:invalid_recipients] ||= []
      recipient = AuthorizableUser.find_by_name(name)

      if skip_recipient_validation
        warnings[:invalid_recipients].push(name)
      elsif recipient.nil? || sender.can_send_to_user?(recipient)
        errors.add(:base, I18n.t('app.models.delivery.send_fail_invalid_recipient_specific', recipient_name: name))
      else
        msg = I18n.t('app.models.delivery.send_fail_permissions_multiple')
        errors.add(:base, msg) unless errors.full_messages.include?(msg)
      end
    end
  end

  def remove_external_flag(name)
    if /(.+)\(external\)/.match(name)
      name = $1.strip
    end

    name
  end

  def use_tos?
    EConfiguration.tos_for_sending.present?
  end

  def check_notifications
    if ActiveRecord::ConnectionAdapters::Column.value_to_boolean @notify_on_upload
      expand_distribution_lists((Utils.parse_recipients(notifiable_on_upload))).each do |name|
        name = remove_external_flag(name)

        if valid_cc? name
          build_notification name, 'upload'
        else
          cc = AuthorizableUser.find_by_name(name)

          if cc.nil? || sender.can_send_to_user?(cc)
            errors.add(:base, I18n.t('app.models.delivery.send_fail_invalid_cc', cc_name: name))
          else
            msg = I18n.t('app.models.delivery.send_fail_invalid_ccs_upload')
            errors.add(:base, msg) unless errors.full_messages.include?(msg)
          end
        end
      end
    end

    if ActiveRecord::ConnectionAdapters::Column.value_to_boolean @notify_on_download
      expand_distribution_lists(Utils.parse_recipients(notifiable_on_download)).each do |name|
        name = remove_external_flag(name)

        if valid_cc? name
          build_notification name, 'download'
        else
          cc = AuthorizableUser.find_by_name(name)

          if cc.nil? || sender.can_send_to_user?(cc)
            errors.add(:base, I18n.t('app.models.delivery.send_fail_invalid_cc', cc_name: name))
          else
            msg = I18n.t('app.models.delivery.send_fail_invalid_ccs_download')
            errors.add(:base, msg) unless errors.full_messages.include?(msg)
          end
        end
      end
    end

    if ActiveRecord::ConnectionAdapters::Column.value_to_boolean @notify_on_receipt
      expand_distribution_lists(Utils.parse_recipients(notifiable_on_receipt)).each do |name|
        name = remove_external_flag(name)
        if valid_cc? name
          build_notification name, 'receipt'
        else
          cc = AuthorizableUser.find_by_name(name)

          if cc.nil? || sender.can_send_to_user?(cc)
            errors.add(:base, I18n.t('app.models.delivery.send_fail_invalid_cc', cc_name: name))
          else
            msg = I18n.t('app.models.delivery.send_fail_invalid_ccs_receipt')
            errors.add(:base, msg) unless errors.full_messages.include?(msg)
          end
        end
      end
    end

  end

  def check_metadata
    return unless metadata_profile.present?
    json_errors = metadata_profile.json_errors(metadata_json)
    json_errors += metadata_profile.title_and_note_errors(self.title, self.note)
    json_errors.each { |e| errors.add(:metadata, e) }
  end

  def user_can_transfer_from_share?(user, share)
    if !user.accessible_shares.include?(share)
      return false
    elsif share.remote? && !user.can_create_from_remote?
      return false
    else
      return true
    end
  end

  def check_sources
    if sources && sources.count != 1
      errors.add(:sources, I18n.t('app.models.delivery.send_fail_must_have_one_source'))
    else
      if source_share.nil?
        errors.add(:sources, I18n.t('app.models.delivery.send_fail_unknown_share_id', share_id: (sources.first[:id])))
      elsif !user_can_transfer_from_share?(sender, source_share)
        errors.add(:sources, I18n.t('app.models.delivery.send_fail_source_not_authorized'))
      end
    end
  end

  ##
  # This silly method is so we can share some of the validation logic
  # with the overlay api.
  def self.do_check_paths(paths)
    unless paths && paths.count > 0
      return I18n.t('app.models.delivery.send_fail_no_items_to_upload')
    end
    basenames = paths.map {|sp| File.basename(sp) }
    unless basenames == basenames.uniq
      return I18n.t('app.models.delivery.send_fail_unique_filenames')
    end

    # All OK; return nil error message
    return nil
  end

  def check_paths
    source_paths = sources.first[:paths]
    error_message = Delivery.do_check_paths(source_paths)
    if error_message
      errors.add_to_base(error_message)
    end
  end

  def check_ear
    if sources.present?
      share = SourceShare.find_by_id(sources.first[:id])
      if share && share.remote? && require_ear? && ear_passphrase.blank? && !share.enable_cloud_referencing?
        errors.add(:ear_passphrase, I18n.t('app.models.delivery.send_fail_ear_must_be_specified'))
      end
    end
    if disallow_ear? && (use_encryption_at_rest == '1' || ear_passphrase.present?)
      errors.add(:ear_passphrase, I18n.t('app.models.delivery.send_fail_ear_not_allowed'))
      self.error_code = ErrorCodes::EAR_NOT_ALLOWED
      self.error_message = I18n.t('app.models.delivery.send_fail_ear_disabled_message')
    end
  end

  def disallow_ear?
    if dropbox_delivery?
      dropbox.disallow_ear?
    elsif EConfiguration.can_override_ear_per_personal_invite?
      !delivery.ear_enabled
    else
      EConfiguration.disallow_ear?
    end
  end

  def sender_share_id_same
    return unless forward? && metadata_hash['SenderShareId'].present?

    unless metadata_hash['SenderShareId'].to_i == original_delivery.default_destination.id
      errors.add_to_base(t('app.models.delivery.send_fail_sendershareid_forwarding'))
    end
  end

public

  def self.search_for(search, options = {})
    page = options.delete(:page) || 1
    joins_list = [:recipient_deliveries]
    if (options[:joins])
      joins_list.push(options[:joins])
      options.delete(:joins)
    end
    paginate({
      page: page,
      joins: joins_list,
      select: 'DISTINCT deliveries.*',
      conditions: [
        'sender_name LIKE :search OR
         title       LIKE :search OR
         note        LIKE :search OR
         recipient_deliveries.recipient_name LIKE :search',
        { search: "%#{search}%" }
      ]
    }.merge(options))
  end

  def self.paginated_pending_list(page, user, order, search)
    paginate(
      :page => page,
      :per_page => user.rows_per_page,
      :conditions => [
        "deliveries.sender_id=? AND " +
        "deliveries.state = ? AND " +
        "deliveries.release_policy != ? AND " +
        "(deliveries.archived=false OR deliveries.archived IS NULL) AND " +
        "e_packages.e_package_status = ? AND " +
        "(e_packages.e_package_files_deleted_by_admin_time IS NULL) AND " +
        "(deliveries.title LIKE ? OR deliveries.note LIKE ? OR recipient_deliveries.recipient_name LIKE ?)",
        user.id,
        Delivery::STATE_HELD,
        Delivery::RELEASE_POLICY_NOW,
        EPackage::STATUS_VALID,
        "%#{search}%",
        "%#{search}%",
        "%#{search}%"
      ],
      :order => order,
      :include => [:e_package, :sender, :recipient_deliveries]
    )
  end

  def self.paginated_pending_history(page, user, order, search)
    paginate(
        :page => page,
        :per_page => user.rows_per_page,
        :conditions => [
            "deliveries.sender_id=? AND " +
                "deliveries.state = ? AND " +
                "deliveries.release_policy != ? AND " +
                "(deliveries.title LIKE ? OR deliveries.note LIKE ? OR recipient_deliveries.recipient_name LIKE ?)",
            user.id,
            Delivery::STATE_HELD,
            Delivery::RELEASE_POLICY_NOW,
            "%#{search}%",
            "%#{search}%",
            "%#{search}%"
        ],
        :order => order,
        :include => [:e_package, :sender, :recipient_deliveries]
    )
  end

  def self.paginated_sent_list(page, user, order, search)
    paginate(
        :page => page,
        :per_page => user.rows_per_page,
        :conditions => [
            "deliveries.sender_id=? AND " +
                "(deliveries.state = ? OR deliveries.release_policy = ?) AND " +
                "(deliveries.archived=false OR deliveries.archived IS NULL) AND " +
                "e_packages.e_package_status = ? AND " +
                "(e_packages.e_package_files_deleted_by_admin_time IS NULL) AND " +
                "(deliveries.title LIKE ? OR deliveries.note LIKE ? OR recipient_deliveries.recipient_name LIKE ?)",
            user.id,
            Delivery::STATE_RELEASED,
            Delivery::RELEASE_POLICY_NOW,
            EPackage::STATUS_VALID,
            "%#{search}%",
            "%#{search}%",
            "%#{search}%"
        ],
        :order => order,
        :include => [:e_package, :sender, :recipient_deliveries]
    )
  end


  def self.paginated_sent_history(page, user, order, search)
    paginate(
      :page => page,
      :per_page => user.rows_per_page,
      :conditions => [
        "deliveries.sender_id=? AND (deliveries.state = ? OR deliveries.release_policy = ?) AND (deliveries.title LIKE ? OR deliveries.note LIKE ? OR recipient_deliveries.recipient_name LIKE ?)",
        user.id, Delivery::STATE_RELEASED, Delivery::RELEASE_POLICY_NOW, "%#{search}%", "%#{search}%", "%#{search}%"
      ],
      :order => order,
      :include => [:e_package, :sender, :recipient_deliveries]
    )
  end

  def lookup_recipients(recipient_list, for_ccs=false)
    recipient_list.map{ |r| lookup_recipient(r, true, for_ccs) || r }
  end

  def sender_email
    u = User.find(sender_id)
    u ? u.email : ''
  end

  def wg_recipient_emails
    recipient_wg_members = []
    workgroup_deliveries.each do |wg|
      recipient_wg_members.concat wg.get_members()
    end
    recipient_wg_members.collect{|x| x.email}
  end

  def recipient_emails
    valid_user_deliveries = user_deliveries.reject{|d| d.recipient.nil?}
    emails = valid_user_deliveries.collect{|d| d.recipient.email }.compact.uniq.join(',')
    emails
  end

  # Assumption: sender is local
  def remote_servers
    this_server = HomeServer.this_server
    relay_servers = []
    each_recipient do |recipient|
      recipient.home_servers.each do |home_server|
        relay_servers << home_server if ((home_server != this_server) && !relay_servers.include?(home_server))
      end
    end
    relay_servers
  end

  def recipients_for_home_server(home_server)
    recipients = []
    each_recipient do |recipient|
      recipients << recipient if recipient.home_servers.include?(home_server)
    end
    recipients
  end

  def external_recipients_for_home_server(home_server)
    recipients = []
    each_recipient do |recipient|
      recipients << recipient if (recipient &&
                                  recipient.is_a?(User) &&
                                  recipient.external_email_user? &&
                                  (recipient.home_server == home_server))
    end
    recipients
  end

  def each_recipient
    self.all_recipients.each do |d|
      if d.first=='*'
        workgroup = EWorkgroup.find_by_e_wg_name(d[1,d.length-1])
        yield workgroup if workgroup
      else
        recipient = User.find_by_name(d)
        yield recipient if recipient
      end
    end
  end

  def remove_from_senders_view
    # remove the delivery only if the package is not uploading
    # (state might have changed since the web browser link was clicked)
    unless self.e_package.busy_uploading?
      self.update_attribute(:archived, true)
    end
  end

  def remove_invalid_recipient_deliveries
    user_deliveries.each do |d|
      if !d.recipient
        logger.warn("Could not send package #{title} to a deleted user")
        d.destroy
      elsif !d.recipient.account_activated
        logger.warn("Could not send package #{title} to user because this user is deactivated")
        d.destroy
      elsif !d.sender.can_send_to_user?(d.recipient)
        logger.warn("Could not send package #{title} to user because sender no longer has permission to send to this user")
        d.destroy
      end
    end

    workgroup_deliveries.each do |d|
      if !d.recipient
        logger.warn("Could not send package #{title} to a deleted workgroup")
        d.destroy
      elsif !d.recipient.can_user_send_to_wg?(d.sender)
        logger.warn("Could not send package #{title} to recipient because sender no longer has permission to send to this workgroup")
        d.destroy
      end
    end
  end

  def send_received_notifications(share)
    notified_users = Set.new
    to_notify = user_deliveries.find_all_by_share_id(share.id) + workgroup_deliveries.find_all_by_share_id(share.id)

    if FASPEX_CONFIG['EnableMultipleEmailsPerShare']
      to_notify.each do |recipient_delivery|
        recipient_delivery.send_received_notifications(notified_users)
      end
    else
      to_notify.each do |recipient_delivery|
        notified_users += recipient_delivery.send_received_notifications(notified_users)
      end
    end
  end

  def send_receipt_cc_notifications
    send_sent_ccs
    send_received_ccs
  end

  def send_received_ccs
    package = EPackage.find self.e_package_id
    recipients.each do |r|
      user = AuthorizableUser.find_by_name(r) || AuthorizableUser.find_by_email(r)
      if user
        user.full_email_cc_list.each do |email|
          BackgroundNotifier.deliver_package_cc_receipt_on_received(user, package, self, {:name => email, :email => email})
        end
      end
    end
  end

  def send_sent_ccs
    cc_on_sent = notifications.on_receipt.empty? ? sender.full_email_cc_list : notifications.on_receipt
    package = EPackage.find self.e_package_id
    cc_on_sent.each do |n|
        notif = n.is_a?(Notification) ? {:name => n.name, :email => n.email } : {:name => n, :email => n }
        BackgroundNotifier.deliver_package_cc_receipt_on_sent(package, self, notif)
    end
  end

  def send_del_dled_notif(downloader, partial_package)
    package = EPackage.find self.e_package_id
    sender  = User.find self.sender_id

    if downloader.notify_downloader_on_download
      package_url = EConfiguration.get_full_main_server_address + received_path(RecipientDelivery.find_by_delivery_id(self.id), local: downloader.force_local_login?)

      # Add SAML Provider id if SAML user
      package_url = Delivery.add_saml_provider_id(package_url, downloader.saml_configuration_id) if downloader.saml?

      BackgroundNotifier.deliver_package_downloaded_notification(downloader, package, self, package_url, partial_package, downloader)
    end


    return if package.full_complete_download_count_by_user(downloader) > 1
    return if package.partial_complete_download_count_by_user(downloader) > 1
    #return if sender.name == downloader.name

    notifications.on_download.each do |n|
      if sender.email == n.email

        package_url = EConfiguration.get_full_main_server_address + sent_path(self.id, local: sender.force_local_login?)

        # Add SAML Provider id if SAML user
        package_url = Delivery.add_saml_provider_id(package_url, sender.saml_configuration_id) if sender.saml?

        BackgroundNotifier.deliver_package_downloaded_notification(downloader, package,
                                                         self, package_url,
                                                         partial_package, sender)
      else
        BackgroundNotifier.deliver_package_downloaded_cc_notification(downloader, package, self, n)
      end
    end
  end

  def has_forwards?
    !Delivery.find_by_forward_and_e_package_id(true, e_package_id).nil?
  end

  def has_relays?
    !self.package_relays.empty?
  end

  def relay_status
    return nil if !has_relays?
    conditions = ["delivery_id = ?", self.id]
    state_min = PackageRelay.minimum(:relay_state, :conditions => conditions)
    state_min
  end

  def pretty_relay_status
    case relay_status
    when PackageRelay::PENDING
      "Relay Pending"
    when PackageRelay::RETRY
      "Retry Relay"
    when PackageRelay::RELAYING
      "Relaying"
    when PackageRelay::FAILED
      "Relay Failed"
    when PackageRelay::COMPLETE
      "Complete"
    end
  end

  def metadata_and_note
    data = ''
    if !user_metadata_fields.empty?
      data += "#{self.metadata_lines}"
    end
    data += "\nNote:\n #{note}" if note and not note.empty?
    data
  end

  def special_metadata_fields
    [
      'SenderShareId',
      'OverrideShareIds',
      'RecipientShareIds'
    ]
  end

  def metadata_lines
    mdata = ""
    user_metadata_fields.each do |key|
      next if special_metadata_fields.include?(key)
      mdata += "#{key}: #{metadata_hash[key]}\n"
    end
    mdata
  end

  def metadata_xml(options = {})
    builder = options[:builder] || Builder::XmlMarkup.new(options)
    builder.metadata do
      metadata = ActiveSupport::JSON.decode(metadata_json)
      if metadata
        metadata.each_pair do |k, v|
          builder.field(v, :name => k)
        end
      end
      builder.field(note, :name => :note)
    end
  end

  def metadata_filename
    "aspera-metadata-#{e_package.upload_transfer_uuid}.xml"
  end

  def save_metadata_to_pkg_dir(share=nil, save_path=nil)
    # Share is different when saving to custom inbox / relay
    share ||= default_destination

    # Path is different when relaying - no package dir structure
    save_path ||= metadata_file_save_path(share)

    share.api_client.create_directory(save_path)

    builder = Builder::XmlMarkup.new(:indent => 2)
    builder.instruct!
    contents = metadata_xml(:builder => builder)

    path = File.join(save_path, metadata_filename)
    share.api_client.create_file(path, contents)
  rescue
    logger.warn("Failed to save metadata XML file for dropbox package #{e_package_id}")
    logger.debug($!.backtrace.join("\n"))
  end

  def metadata_file_save_path(share)
    if FASPEX_CONFIG['SaveMetadataInPackage']
      full_package_path(share)
    else
      outer_package_path(share)
    end
  end

  def upload_from_source
    log_msg = "Uploading package (id: #{e_package_id}) from source share #{e_package.source_share.name}"
    logger.debug(log_msg)
    source_paths = e_package.source_share.node_paths(e_package.e_package_paths.map(&:e_uploader_local_path))
    transfer_job = create_upload_job(e_package.source_share, source_paths, make_symlinks_for_default_destination?, make_cloud_references?)
    symlinked_size = transfer_job.perform
    if make_symlinks_for_default_destination? || make_cloud_references?
      create_fake_upload(e_package.source_share, default_destination, ESession::DIRECTION_CREATE, symlinked_size)
    end
  rescue => e
    logger.warn("#{log_msg} failed")
    logger.debug(e)
    logger.debug(e.backtrace.join("\n"))
    raise EPackage::UploadFromSourceFailed, "#{log_msg} failed"
  end

  def node_path(relative_path)
    sanitized_path = Utils.expand_path(relative_path, '/')
    File.join(inner_directory_path, sanitized_path)
  end

  def node_paths(paths_in_pkg)
    paths_in_pkg.map {|path| node_path(path)}
  end

  def dropbox
    recipient = nil
    if workgroup_deliveries.length == 1
      recipient = workgroup_deliveries.first.recipient
    elsif @recipient_hashes && @recipient_hashes.length == 1
      recipient = @recipient_hashes.keys.first
    end

    if recipient.is_a?(EWorkgroup) && recipient.dropbox?
      recipient
    else
      nil
    end
  end
  memoize :dropbox

  def dropbox_delivery?
    invalid_recipient_names.blank? && dropbox.present?
  end

  def metadata_profile
    if dropbox_delivery?
      MetadataProfile.find_by_id(workgroup_deliveries.first.recipient.dbox_mdata_profile_id)
    else
      MetadataProfile.find_by_id(EConfiguration.first.normal_pkg_mdata_profile_id)
    end
  end
  memoize :metadata_profile

  def save_metadata?
    return false if forward?
    if dropbox_delivery?
      workgroup_deliveries.first.recipient.dbox_save_metadata_to_file?
    else
      EConfiguration.first.save_metadata_to_file?
    end
  end

  def delivery
    self
  end

  def completed_at
    if e_package.completed?
      [e_package.e_package_upload_completion_time, created_at].max
    else
      nil
    end
  end

  def generate_sequence_id
    self.delivery_completion = DeliveryCompletion.create # generate next sequence_id
  end

  def generate_sequence_id!
    generate_sequence_id
    save!
  end

  def storage_share
    default_destination
  end

  def destination_shares
    share_ids = recipient_deliveries.map(&:share_id).uniq - [default_destination.id]
    shares = share_ids.map {|share_id| Share.find_by_id(share_id)}
    shares.reject(&:nil?)
  end

  def all_recipient_shares
    @_recipient_shares ||=
      self.destination_shares +
        self.e_package.all_workgroup_forwardings(self).map(&:share)
  end

  # Replicate to all custom inboxes, override shares and recipient shares
  def replicate_contents
    metadata = JSON.parse(self.metadata_json)

    share_list = destination_shares
    override_shares = self.parse_json_field(metadata['OverrideShareIds']) || []
    recipient_shares = self.parse_json_field(metadata['RecipientShareIds']) || {}
    share_list += override_shares.map{ |id| Share.find_by_id(id) }
    share_list += recipient_shares.map{ |user, id| Share.find_by_id(id) }
    share_list.uniq.map{ |s| relay_to_inbox(s, true) rescue [s.id, $!.message]}
  end

  def relay_to_inboxes
    destination_shares.each {|dest_share| relay_to_inbox(dest_share) }
    metadata_array = ActiveSupport::JSON.decode(self.metadata_json)
    if nil != metadata_array and metadata_array['OverrideShareIds'].present?
      share_array = self.parse_json_field(metadata_array['OverrideShareIds'])
      share_array.each do |share_id|
        begin
          user_share = Share.find(share_id)
          if user_share
            relay_to_inbox(user_share)
          else
            logger.warn("Package routing: overrideShareIds field of metadata does not contain share ID.")
          end
        rescue
          logger.warn("Share ID: #{share_id.to_s} is invalid, skipping")
        end
      end
    end
  end

  def full_package_path(share)
    File.join(share.directory, File.join(e_package.outer_directory_name, e_package.inner_directory_name))
  end

  def outer_package_path(share)
    File.join(share.directory, e_package.outer_directory_name)
  end

  # Will return workgroup if there is one recipient and it's a workgroup
  def single_workgroup_recipient
    recipient = if recipients.length == 1
                  lookup_recipient(recipients.first)
                elsif @recipient_hashes.present? && @recipient_hashes.length == 1
                  @recipient_hashes.keys.first
                end

    recipient if recipient.is_a?(EWorkgroup)
  end

  # Only skip inbox if workgroup is sole recipient
  def workgroup_skip_inbox?
    workgroup = single_workgroup_recipient
    workgroup.present? && workgroup.upload_directly_to_custom_inbox
  end

  # Determine if there is a custom inbox and workgroup is only recipient
  def workgroup_with_custom_inbox?
    workgroup = single_workgroup_recipient
    workgroup.present? && workgroup.destination_share_id.present?
  end

  def tag_hash(download=false, downloader=nil, delivery=nil, dest_share=nil)
    max_tag_length = 1500

    if FASPEX_CONFIG['MaxTagsLength'] && FASPEX_CONFIG['MaxTagsLength'].to_i <= 4096
      max_tag_length = FASPEX_CONFIG['MaxTagsLength'].to_i
    end
    # Priority 1: Add number of recipients.
    tags = {
        :aspera => {
            :faspex => {
                :num_recipients => recipients.length
            }
        }
    }

    # Priority 2: Specify relay shares.
    if !delivery.nil? && !dest_share.nil?
      tags[:aspera][:faspex][:relay] = {
          source_share: {
              id: delivery.default_destination.id,
              name: delivery.default_destination.name,
              node: "#{delivery.default_destination.node.name} (#{delivery.default_destination.node.host})"
          },
          destination_share: {
              id: dest_share.id,
              name: dest_share.name,
              node: "#{dest_share.node.name} (#{dest_share.node.host})"
          }
      }
    end

    # Confirm first two priorities fit.
    if Base64.encode64(JSON.dump(tags)).length > max_tag_length
      tags[:aspera][:faspex].delete(:relay)
      return tags
    end

    # Priority 3: Specify dropbox/non-dropbox workgroup recipients.
    tags[:aspera][:faspex][:dropbox] = (lookup_recipient(recipients.first).e_wg_name rescue nil) if dropbox_delivery?
    workgroups = lookup_recipients(recipients).select{ |r| r.is_a?(EWorkgroup) && !r.dropbox? }
    tags[:aspera][:faspex][:workgroups] = workgroups.map{ |wg| wg.e_wg_name } if workgroups.present?

    # Confirm third priority fits.
    if Base64.encode64(JSON.dump(tags)).length > max_tag_length
      tags[:aspera][:faspex].delete(:dropbox)
      tags[:aspera][:faspex].delete(:workgroups)
      return tags
    end

    # Priority 4: Specify downloader information.
    if download
      tags[:aspera][:faspex][:downloader] = user_params(downloader)

      if Base64.encode64(JSON.dump(tags)).length > max_tag_length
        tags[:aspera][:faspex].delete(:downloader)
        return tags
      end
    end

    # Priority 5: Attempt to add sender params.
    if FASPEX_CONFIG['UserFieldsInTags']
      tags[:aspera][:faspex][:requestor] = user_params(sender)

      if Base64.encode64(JSON.dump(tags)).length > max_tag_length
        tags[:aspera][:faspex].delete(:requestor)
        return tags
      end
    end

    # Priority 6: Add metadata.
    # FASPX-1836 - Changing original_delivery.metadata_json to metadata_json of current delivery.
    if metadata_json.present?
      tags[:aspera][:faspex][:metadata] = JSON.parse(metadata_json)
    end

    if Base64.encode64(JSON.dump(tags)).length > max_tag_length
      tags[:aspera][:faspex].delete(:metadata)
      return tags
    end

    # Priority 7: Add as many recipients as possible if config set.
    if FASPEX_CONFIG['UserFieldsInTags']
      recipient_list = []
      for recipient in recipients
        if !recipient.is_a?(EWorkgroup)
          tags[:aspera][:faspex][:recipients] = recipient_list.push(user_params(lookup_recipient(recipient)))
          if Base64.encode64(JSON.dump(tags)).length > max_tag_length
            tags[:aspera][:faspex][:recipients].pop
            return tags
          end
        end
      end
    end

    tags
  end

  def relay_to_inbox(dest_share, force = false)
    recipient_delivery = recipient_deliveries.first(conditions: {share_id: dest_share.id}) || recipient_deliveries.first
    if recipient_delivery.storage_share == dest_share && recipient_delivery.inner_directory_exists? && !force
      if forward?
        create_fake_upload_forward_and_relay(dest_share)
      else
        send_received_notifications(dest_share)
      end
    else
      default_destination_share = default_destination
      direction = ESession::DIRECTION_RELAY
      source_node = default_destination_share.node
      source_paths = [outer_package_path(default_destination_share)]
      dest_path = dest_share.directory
      cookie = recipient_delivery.prepare_transfer(direction, ['/'], sender, {share: dest_share}).cookie

      # Symlinking
      if partial_symlink = symlink = make_symlinks_for_custom_inbox?(e_package.source_share, dest_share)
        # Source files and custom inbox are on same node - Can symlink package files

        # Change source from default inbox to original source
        source_node = e_package.source_share.node
        source_paths = e_package.source_share.node_paths(e_package.e_package_paths.map(&:e_uploader_local_path))
        # Must build full package path because we're not copying package dir from default inbox
        dest_path = full_package_path(dest_share)
      else
        # Default inbox and custom inbox are on same node - Can symlink entire package
        symlink = make_symlinks_for_custom_inbox?(default_destination_share, dest_share)
      end

      tags = tag_hash
      tags[:aspera][:faspex][:relay] = {
        source_share: {
          id: default_destination_share.id,
          name: default_destination_share.name,
          node: "#{default_destination_share.node.name} (#{default_destination_share.node.host})"
        },
        destination_share: {
          id: dest_share.id,
          name: dest_share.name,
          node: "#{dest_share.node.name} (#{dest_share.node.host})"
        }
      }

      transfer_params = TransferParams.new(
        cookie: cookie,
        target_rate_kbps: select_relay_bandwidth(default_destination_share, dest_share),
        paths: [{destination: dest_path}],
        tags: tags
      )

      symlinked_size = TransferJob.new(source_node, source_paths, dest_share.node, dest_path, transfer_params, {symlink: symlink, overwrite:"Diff"}).perform

      if partial_symlink && self.save_metadata?
        # Only added symlinked files -- metadata file was not copied. Must copy it over
        save_metadata_to_pkg_dir(dest_share)
      end

      create_fake_upload(default_destination_share, dest_share, ESession::DIRECTION_RELAY, symlinked_size) if symlink
    end
  rescue
    Rails.logger.error "Relay to share #{dest_share.id} failed for delivery #{self.id}: #{$!.message}"
  end


  def forward_template (ignore_metadata=false,ignore_note=false)
    result = "\n\n"
    if user_metadata_fields.any? && !ignore_metadata
      result += "--------- Forwarded metadata ---------\n"
      result += delivery.user_metadata_fields.map { |key| "#{key}: #{delivery.metadata_hash[key]}" }.join("\n")
      result += "\n"
    end
    unless note.nil? || ignore_note
      result += "--------- Forwarded note ---------\n"
      result += note
    end
    result
  end

  def make_cloud_references?
    return false if sources.nil?
    source_share = SourceShare.find_by_id(sources.first[:id])

    return false if source_share.client?

    source_share_node = source_share.node
    default_destination_node = default_destination.node

    @make_cloud_references ||= (source_share.uses_azure_storage? || source_share.uses_s3_storage?) &&
        source_share.enable_cloud_referencing? &&
        ( source_share_node.id == default_destination_node.id)
  end

  def pkg_contains_directories?
    source_share = SourceShare.find_by_id(sources.first[:id])
    source_paths = source_share.node_paths(sources.map{ |s| s['paths']}.flatten)
    source_paths.any? do  |source_path|
      source_share.node.api_client.files_info(source_path).response_parsed['paths'][0]['type'] == 'directory'
    end
  end

  # If a package has multiple deliveries, then all after the first are forwards - use original delivery for share / path info.
  def original_delivery
    e_package.deliveries.first || self rescue self
  end

  def is_original_delivery?
    self == original_delivery
  end

  def sender_share_override
    mdata = metadata || JSON.parse(metadata_json) rescue {}
    Share.find_by_id(mdata['SenderShareId'])
  end

  def initial_destination
    # if package exists, return actual share where it is located
    # this case occurs when the package is browsed
    return Share.find_by_id(e_package.storage_share_id) if e_package

    # this case handles uploads to a custom inbox/ SenderShareId metadata
    custom_share = (workgroup_skip_inbox? && single_workgroup_recipient.destination_share) || sender_share_override
    return custom_share if custom_share

    # this case is for normal uploads to the default inbox
    Share.default_destination
  end

  def default_destination(ignore_original = false)
    if is_original_delivery? || ignore_original
      initial_destination
    else
      original_delivery.default_destination rescue e_package.storage_share || Share.default_destination
    end
  end

  def default_download_location(ignore_original = true)
    #if the original delivery had a SenderShareId, force default_destination to return original_delivery.default_destination
    if original_delivery.metadata_hash['SenderShareId'].present?
      default_destination(false)
    else
      default_destination(ignore_original)
    end
  end

  def default_destination_path(ignore_original = false)
    if is_original_delivery? || ignore_original
      e_package.node_path('/')
    else
      original_delivery.default_destination_path
    end
  end

  def default_transfer_host(request, upload=false)
    transfer_share = upload ? default_destination : default_download_location
    transfer_share.node.get_server_address(request.remote_ip, request.host)
  end

  def create_fake_upload(source, dest, direction = ESession::DIRECTION_CREATE, symlinked_size = nil)
    share = dest
    stats = default_destination.node.api_client.du(full_package_path(default_destination)).response_parsed
    total_size = (symlinked_size != nil) ? symlinked_size : stats['total_size']
    file_count = stats['file_count']
    now = Time.now

    transfer_request = TransferRequest.find_by_delivery_id_and_share_id(self.id, share.id)

    sess = ESession.create(
        bytes_pretransfer: total_size,
        bytes_transferred: total_size,
        bytes_written: total_size,
        central_id: SecureRandom.uuid,
        delivery: e_package.deliveries.first,
        delivery_id: self.id,
        direction: direction,
        e_package: e_package,
        e_session_start_notification_time: now,
        e_session_stop_notification_time: now,
        error_code: 0,
        files_completed: file_count,
        files_failed: 0,
        initiator: e_package.sender,
        remote_host: source.node.host,
        share: share,
        status: ESession::STATUS_COMPLETE,
        system_user_name: '',
        transfer_request: transfer_request
    )

    e_package.update_from_session(sess)
  end

  def create_fake_upload_forward_and_relay (dest)
    stats = default_destination.node.api_client.du(full_package_path(default_destination)).response_parsed
    total_size = stats['total_size']
    file_count = stats['file_count']
    now = Time.now
    transfer_cookie = TransferCookie.for_transfer(ESession::DIRECTION_RELAY, sender, self)

    transfer_request = TransferRequest.create(
        cookie: transfer_cookie.to_s,
        delivery: self,
        direction: ESession::DIRECTION_RELAY,
        e_package: self.e_package,
        flatten: false,
        partial_package: false,
        share: dest,
        user: sender,
        uuid: transfer_cookie.transfer_uuid
    )

    sess = ESession.create(
        bytes_pretransfer: total_size,
        bytes_transferred: total_size,
        bytes_written: total_size,
        central_id: SecureRandom.uuid,
        delivery: self,
        delivery_id: self.id,
        direction: ESession::DIRECTION_RELAY,
        e_package: e_package,
        e_session_start_notification_time: now,
        e_session_stop_notification_time: now,
        error_code: 0,
        files_completed: file_count,
        files_failed: 0,
        initiator: sender,
        initiator_name: sender.name,
        initiator_user_id: sender.id,
        remote_host: default_destination.node.host,
        share: dest,
        status: ESession::STATUS_COMPLETE,
        system_user_name: '',
        transfer_request: transfer_request
    )

    e_package.update_from_session(sess)
  end

  def select_relay_bandwidth(source_share, destination_share)
    if source_share.override_transfer_parameters? && destination_share.override_transfer_parameters?
      [source_share.relay_outgoing_bandwidth, destination_share.relay_incoming_bandwidth].min
    elsif source_share.override_transfer_parameters?
      [CConfiguration.first.c_relay_outgoing_bandwidth, source_share.relay_outgoing_bandwidth].min
    elsif destination_share.override_transfer_parameters?
      [CConfiguration.first.c_relay_outgoing_bandwidth, destination_share.relay_incoming_bandwidth].min
    else
      CConfiguration.first.c_relay_outgoing_bandwidth
    end
  end

private

  def valid_cc? name
    # Strip (external) flag

    Utils.valid_email?(name) || sender.lookup_user_recipient(name)
  end

  def build_notification name, event
    if user = AuthorizableUser.find_by_name(name)
      notifications.build name: user.full_name, email: user.email, event: event
    else
      notifications.build name: name, email: name, event: event
    end
  end

  def make_symlinks_for_custom_inbox?(source, dest)
    @make_symlinks ||= source.enable_linking &&
        source.node == dest.node &&
        ear_passphrase.nil? &&
        source.node.supports_symlinks?
  end

  def make_symlinks_for_default_destination?
    @make_symlinks ||= e_package.source_share.enable_linking? &&
        e_package.source_share.node == default_destination.node &&
        ear_passphrase.nil? &&
        e_package.source_share.node.supports_symlinks?
  end

  def create_upload_job(source_share, source_paths, symlink = false, cloud_reference = false)
    #source_paths = e_package.source_share.node_paths(e_package.e_package_paths.map(&:e_uploader_local_path))
    cookie = prepare_transfer(ESession::DIRECTION_CREATE, source_paths, sender).cookie
    transfer_params = TransferParams.new(
      cookie: cookie,
      target_rate_kbps: select_relay_bandwidth(source_share, default_destination),
      tags: tag_hash
    )
    expanded_source_paths = []
    if EConfiguration.require_obfuscation? || self.obfuscation_enabled
      pairs = []

      source_paths.map do |source_path|
        recursive_obfuscation source_share, source_path, pairs, expanded_source_paths
      end
      transfer_params.paths = pairs
    else
      transfer_params.paths = [{destination: default_destination_path}]
    end
    final_source_paths = (EConfiguration.require_obfuscation? || self.obfuscation_enabled) ? expanded_source_paths : source_paths
    TransferJob.new(
      source_share.node,
      final_source_paths,
      default_destination.node,
      default_destination_path,
      transfer_params,
      ear_passphrase: ear_passphrase,
      symlink: symlink,
      cloud_reference: cloud_reference
    )
  end

  def recursive_obfuscation(source_share, source_path, pairs, file_paths = [], expanded_source_paths)
    #Default return size for browse node api call is 100 items,
    # for folders with >100 files, we hardcode to 10000 and set as known limitation
    node_api_response = source_share.node.api_client.browse(source_path, :count => 10000)
        directory_items = node_api_response.directory_items
        if directory_items.length == 0
          file_name = source_path.split("/").last
          expanded_source_paths << source_path
          pairs <<
          {
            'source' => source_path,
            'destination' =>  obfuscated_destination_path(file_paths, file_name).gsub(/=/, "")
          }
        else
          file_paths << source_path.split("/").last
          directory_items.each do |directory_item|
            if directory_item.type == "directory"
              recursive_obfuscation source_share, directory_item.path, pairs, file_paths, expanded_source_paths
            else
              file_name = directory_item.path.split("/").last
              expanded_source_paths << directory_item.path
              pairs <<
              {
                'source' => directory_item.path,
                'destination' =>  obfuscated_destination_path(file_paths, file_name).gsub(/=/, "")
              }
            end
          end
        end
  end


  def obfuscated_destination_path(file_paths, file_name)
    obfuscated_paths_array = []
    obfuscated_file_name = obfuscate_filename file_name
    file_paths.each do |path|
      obfuscated_paths_array << ActiveSupport::Base64.urlsafe_encode64(path)
    end
    obfuscated_full_path = File.join(obfuscated_paths_array.join("/"), obfuscated_file_name)
    default_destination_path + obfuscated_full_path
  end

  def self.add_saml_provider_id(package_url, saml_id)
    if package_url.include?('?')
      package_url += "&saml_provider=#{saml_id}"
    else
      package_url += "?saml_provider=#{saml_id}"
    end

    package_url
  end

  # ====================================================================== #
  # ====================================================================== #
  #                                                                        #
  # FASPEX OVERLAY API DEFINIIONS AND METHODS                              #
  #                                                                        #
  # ====================================================================== #
  # ====================================================================== #

  public

  ALLOWED_SORT_FIELDS = %w(
    download_limit
    expiration_date
    release_date
    title
    updated_at
  )

  ALLOWED_PACKAGE_SORT_FIELDS = %w(
    e_package_contents_exist
    e_package_total_files
    e_package_total_size_bytes
  )

  DEFAULT_RECORDS_PER_PAGE = 10
  #
  # Release Policies
  # ----------------
  #
  RELEASE_POLICY_NOW     = 'release_now'
  RELEASE_POLICY_ON_DATE = 'release_on_date'
  RELEASE_POLICY_NEVER   = 'do_not_release'
  #
  # Primary States
  # --------------
  #
  # IMPORTANT: If you add/remove states, make sure you update #check_state.
  #
  STATE_HELD     = 'held'
  STATE_RELEASED = 'released'
  STATE_EXPIRED  = 'expired'
  STATE_DELETED  = 'deleted'
  #
  # Transient States
  # ----------------
  #
  # These are states a delivery can be in while the server is processing
  # a transition. For example, when a delivery transitions from held to live,
  # the server has to initiate package routing in the background; this is
  # non-blocking, and the server keeps checking until all of that is done; then
  # it will be safe to mark the delivery as live.
  #
  # IMPORTANT: If you add/remove states, make sure you update #check_state.
  #
  TRANSIENT_STATE_HELD_TO_RELEASED_START = 'transient_held_to_released_start'
  TRANSIENT_STATE_HELD_TO_RELEASED_ROUTING_PACKAGES =
      'transient_held_to_released_routing_packages'
  TRANSIENT_STATE_HELD_TO_RELEASED_FINISH = 'transient_held_to_released_finish'
  TRANSIENT_STATE_HELD_TO_EXPIRED = 'transient_held_to_expired'
  TRANSIENT_STATE_HELD_TO_DELETED = 'transient_held_to_deleted'
  TRANSIENT_STATE_RELEASED_TO_EXPIRED = 'transient_released_to_expired'
  TRANSIENT_STATE_RELEASED_TO_DELETED = 'transient_released_to_deleted'
  TRANSIENT_STATE_EXPIRED_TO_DELETED = 'transient_expired_to_deleted'
  #
  # Mailboxes
  # ---------
  #
  MAILBOX_INBOX = 'inbox'
  MAILBOX_OUTBOX = 'outbox'
  #
  # Transfer directions
  # -------------------
  #
  # (These are slightly different from ESession directions.)
  #
  TRANSFER_DIRECTION_SEND = 'send'
  TRANSFER_DIRECTION_RECEIVE = 'receive'
  #
  # API-related fields, defaults, and validation
  # --------------------------------------------
  #
  attr_accessible :release_policy,
                  :release_date,
                  :expiration_date,
                  :download_limit,
                  :archived,
                  :state

  default_value_for :release_policy, RELEASE_POLICY_NOW
  default_value_for :state, Delivery::STATE_HELD

  validate_on_create :check_release_policy
  validate :check_state

  # ======================================================================
  # METHODS
  # ======================================================================

  # ----------------------------------------------------------------------
  # Validation
  # ----------------------------------------------------------------------
  #
  # These are ran before every create (database insert) or update to ensure
  # we have valid data in the db.

  ##
  # Validates that the delivery's #state is one of the Delivery::STATE_*
  # or Delivery::TRANSIENT_STATE_* constants.
  #
  # IMPORTANT: If you add/remove states, make sure you update this method.
  def check_state
    #
    # State must be one of these values
    #
    unless [STATE_HELD,
            STATE_RELEASED,
            STATE_EXPIRED,
            STATE_DELETED,
            TRANSIENT_STATE_HELD_TO_RELEASED_START,
            TRANSIENT_STATE_HELD_TO_RELEASED_ROUTING_PACKAGES,
            TRANSIENT_STATE_HELD_TO_RELEASED_FINISH,
            TRANSIENT_STATE_HELD_TO_EXPIRED,
            TRANSIENT_STATE_HELD_TO_DELETED,
            TRANSIENT_STATE_RELEASED_TO_EXPIRED,
            TRANSIENT_STATE_RELEASED_TO_DELETED,
            TRANSIENT_STATE_EXPIRED_TO_DELETED,
    ].include?(self.state)
      errors.add(:base, "Invalid state: #{self.state}")
    end
  end

  ##
  # Ensures release policy is one of release_now, release_never, or
  # release_on_date.
  def check_release_policy
    unless self.release_policy
      #
      # Default to release now
      #
      self.release_policy = RELEASE_POLICY_NOW
    end
    unless [RELEASE_POLICY_NEVER,
            RELEASE_POLICY_NOW,
            RELEASE_POLICY_ON_DATE].include?(self.release_policy)
      errors.add(:base, I18n.t('app.models.delivery.send_fail_invalid_release_policy', rp: self.release_policy))
    end

    if self.release_policy == RELEASE_POLICY_ON_DATE
      errors.add(:base, I18n.t('app.models.delivery.send_fail_policy_date_future')) if self.release_date.nil? || self.release_date <= Time.now
    end

    if self.release_policy == RELEASE_POLICY_NOW
      self.release_date = Time.now
    end


    errors.add(:base, I18n.t('app.models.delivery.send_fail_expiration_date_future')) if self.expiration_date.present? && self.expiration_date < Time.now

    # self.release_date = self.date_to_iso_string_format(Time.now) if nil == self.release_date and self.release_policy == RELEASE_POLICY_NOW
  end

  # ----------------------------------------------------------------------
  # Create
  # ----------------------------------------------------------------------

  # Errors
  class CreateError < RuntimeError; end

  ##
  # If maybe_date is nil, returns nil. Otherwise, parses maybe_date (this may
  # throw an exception).
  def self.parse_date(maybe_date)
    if maybe_date
      DateTime.parse(maybe_date)
    else
      nil
    end
  end

  def parse_json_field(value)
    Delivery.parse_json_field(value)
  end

  def self.parse_json_field(value)
    value.is_a?(Hash) ? value : value.blank? ? nil : JSON.parse(value)
  end

  ##
  # Convert to legacy faspex notifiable list. arr should be an array of
  # strings, or nil. If arr is nil, returns empty string.
  #
  # For some reason, legacy faspex wants the list of notifiable users
  # as a string like this:
  #
  #    ' "user1name", "user2name" '
  #
  # See check_notifications.
  def self.convert_array_to_legacy_faspex_notifiable_list(arr)
    #
    # Return empty string if arr is nil or empty.
    #
    unless arr && arr.length > 0
      return ""
    end
    #
    # Convert arr to string.
    #
    out = arr.inject("") do |result, name|

      result + "\"#{name}\","

    end
    #
    # Return string with trailing comma removed.
    #
    out.chop
  end

  ##
  # Creates new delivery and e_package with creator as the sender. Uses
  # hash in api_request_body_hash to build them. The hash should follow
  # the Faspex Overlay API.
  #
  # Raises CreateError if values in hash are invalid (e.g. bad dates).
  def self.for_user_and_hash(creator, api_request_body_hash)
    #
    # If this is an external user creating via invitation, they
    # can only send to the inviter. For example, if the 'inviter' is
    # a dropbox, the sender can only send to the dropbox.
    #
    invitation = Invitation.invitation_from_passcode(
        api_request_body_hash[:passcode])
    if invitation
      api_request_body_hash[:recipients] = [invitation.inviter.addressable_name]
    end
    #
    # Sanitize metadata
    #
    if invitation &&
        invitation.inviter.is_a?(EWorkgroup) &&
        invitation.inviter.dropbox?
      maybe_dropbox = invitation.inviter
    else
      # Determine if recipient is dropbox
      recipients = (api_request_body_hash[:recipients].is_a? String) ? JSON.parse(api_request_body_hash[:recipients]) : api_request_body_hash[:recipients]

      if recipients && recipients.length == 1
        recipient = creator.lookup_recipient(recipients.first)
        maybe_dropbox = recipient if recipient && recipient.is_a?(EWorkgroup) && recipient.dropbox?
      end
    end
    metadata_hash =
        MetadataProfile.sanitize_metadata_array_using_active_profile(
            api_request_body_hash[:metadata],
            maybe_dropbox)
    #
    # Sanitize / parse other fields
    #
    begin
      api_request_body_hash[:release_date] =
          self.parse_date(api_request_body_hash[:release_date])
    rescue
      raise CreateError,
            "Bad release date: <#{api_request_body_hash[:release_date]}>"
    end
    begin
      api_request_body_hash[:expiration_date] =
          self.parse_date(api_request_body_hash[:expiration_date])
    rescue
      raise CreateError,
            "Bad expiration date: <#{api_request_body_hash[:expiration_date]}>"
    end
    api_request_body_hash[:notifiable_on_upload] =
        self.convert_array_to_legacy_faspex_notifiable_list(
            api_request_body_hash[:notifiable_on_upload])
    api_request_body_hash[:notifiable_on_download] =
        self.convert_array_to_legacy_faspex_notifiable_list(
            api_request_body_hash[:notifiable_on_download])
    #
    # Create new Delivery record (.new does not save to db yet)
    #
    # We use an explicit constructor so that the creator can't accidentally
    # or maliciously set a value that should be internal / read-only.
    #
    # (We have to use .new so we can configure validation - see below.)
    #
    delivery = Delivery.new(:sender => creator)
    delivery.attributes = {
        :title => api_request_body_hash[:title],
        :note => api_request_body_hash[:note],
        :sender => creator,
        :recipients => recipients,
        :private_recipients => api_request_body_hash[:private_recipients],
        :metadata => metadata_hash,
        :release_policy => api_request_body_hash[:release_policy],
        :release_date => api_request_body_hash[:release_date] || Time.now,
        :expiration_date => api_request_body_hash[:expiration_date],
        :archived => api_request_body_hash[:archived],
        :download_limit => api_request_body_hash[:download_limit],
        :invitation => invitation,
        :ear_passphrase => api_request_body_hash[:ear_passphrase],
        :ear_enabled => api_request_body_hash[:ear_enabled],
        :notify_on_upload => api_request_body_hash[:notify_on_upload],
        :notify_on_download => api_request_body_hash[:notify_on_download],
        :notifiable_on_upload => api_request_body_hash[:notifiable_on_upload],
        :notifiable_on_download =>
            api_request_body_hash[:notifiable_on_download],
        :skip_recipient_validation => api_request_body_hash[:skip_recipient_validation] || false
    }

    if api_request_body_hash[:forward]
      original_delivery = api_request_body_hash[:original_delivery]
      delivery.e_package = original_delivery.e_package
      delivery.forward = true
    end

    #
    # With the Faspex Overlay API, the creator *does not* specify sources
    # or paths. They will initiate a separate request to upload package
    # content.
    #
    # So, we skip source and path validation.
    #
    delivery.skip_source_validation = true
    delivery.skip_path_validation = true
    #
    # Try to save / create record
    #
    if delivery.save
      return delivery
    else
      raise CreateError,
            delivery.formatted_error_messages
    end
  end

  ##
  # Creates directory on the delivery's storage share for the
  # package contents.
  #
  # Raises CreateError if this fails.
  def create_storage_directory
    #
    # Use Node API to create directory
    #
    directory_path = self.storage_directory_path
    response = self.storage_share.api_client.create_directory(
        directory_path)
    #
    # Ensure all is OK.
    #
    if response.error? && !response.user_message.include?('Path already exists')
      raise CreateError,
            "Failed to create directory on package's storage share
#{self.storage_share.name}. The directory we tried to
create was #{self.storage_directory_path}. The error message from the
Node API is #{response.user_message}"
    end
  end

  # ----------------------------------------------------------------------
  # Forward
  # ----------------------------------------------------------------------

  # Errors
  class ForwardBadParamsError < RuntimeError; end
  class ForwardServerError < RuntimeError; end

  def check_forward_permissions(requester)
    unless self.e_package.can_be_forwarded_by?(requester)
      raise ForwardBadParamsError,
            "You (#{requester.name}) are not authorized to forward " +
                "this package (title: #{self.title})."
    end
  end

  def create_forward(requester, api_request_hash)
    # Don't allow forwarding of non released packages
    if self.release_policy == RELEASE_POLICY_NEVER ||
        self.release_policy == RELEASE_POLICY_ON_DATE &&
            self.release_date > Time.now
      raise Api::BadRequestError, "Unreleased packages may not be forwarded"
    end

    #
    # Ensure requester can forward
    #
    self.check_forward_permissions(requester)
    #
    # Do full create of new db records. Append forward template to note.
    #
    begin
      ignore_metadata = (api_request_hash[:exclude_original_metadata].present? && api_request_hash[:exclude_original_metadata].to_s != "false" ) || false
      ignore_note =  (api_request_hash[:exclude_original_note].present? && api_request_hash[:exclude_original_note].to_s != "false") || false
      api_request_hash[:note] = api_request_hash[:note].to_s + self.forward_template(ignore_metadata,ignore_note)
      api_request_hash[:forward] = true
      api_request_hash[:original_delivery] = self
      new_delivery = Delivery.for_user_and_hash(requester, api_request_hash)
    rescue CreateError => e
      raise ForwardBadParamsError,
            e.message
    rescue
      raise ForwardBadParamsError,
            $!.message
    end

    new_delivery.update_attribute(:state, STATE_HELD)
    #
    # Coast is clear - The background processes will monitor the content
    # copying and mark the new delivery as held when that completes. It will
    # then go through the normal life cycle from there.
    #
    # Return new delivery
    #
    new_delivery
  end

  # ----------------------------------------------------------------------
  # Delete
  # ----------------------------------------------------------------------

  # Errors
  class DeleteError < RuntimeError; end

  def do_delete_from_share(share)
    #
    # Determine path to package on share.
    #
    path = share.node_path(self.e_package.outer_directory_name)
    #
    # Use Node API to do delete. (This code is adapted from
    # EPackage#delete_contents.)
    #
    response = share.node.api_client.delete([path])
    if (response.error? &&
        (response.path.nil? || response.path['error']['code'] != 404))
      raise Api::ServerError,
            ('Error deleting package with id ' +
                "#{self.e_package.id} and path #{path} on " +
                "share with id #{share.id}. Message from share: " +
                "#{response.user_message}")
    end
  end

  def do_delete_from_shares(shares)
    shares.each do |share|
      self.do_delete_from_share(share)
    end
  end

  def delete_from_recipient_shares
    #
    # Look up all of the (recipient) shares where the package is stored.
    #
    #
    # Don't delete from the sender / creator's share. (This is equivalent
    # to the delivery's storage share.)
    #
    all_recip_shares = all_recipient_shares - [self.storage_share]
    #
    # Do deletes.
    #
    self.do_delete_from_shares(all_recip_shares)
  end

  def do_full_delete
    #
    # Put delivery in DELETING state. We do a transaction
    # to ensure this is atomic.
    #
    Delivery.transaction do
      #
      # Reload in case someone changed the delivery's state.
      #
      self.reload
      #
      # Update delivery state to DELETING, based on current state.
      #
      case self.state

        when Delivery::STATE_HELD
          self.update_state(Delivery::TRANSIENT_STATE_HELD_TO_DELETED)
        when Delivery::STATE_RELEASED
          self.update_state(Delivery::TRANSIENT_STATE_RELEASED_TO_DELETED)
        when Delivery::STATE_EXPIRED
          self.update_state(Delivery::TRANSIENT_STATE_EXPIRED_TO_DELETED)
        when Delivery::STATE_DELETED
          # nothing to do; just return.
          return
        else
          #
          # Someone else is trying to update the delivery; reject
          # request.
          #
          raise DeleteError,
                ('You cannot delete this package, someone else is trying' +
                    " to update it. (Package id #{self.id}.)")
      end
    end
    #
    # Delete package contents from recipient shares.
    #
    self.delete_from_recipient_shares
    #
    # Mark state as DELETED and mark underlying e_package for deletion.
    #
    self.mark_as_deleted
    self.e_package.mark_for_deletion
  end

  def full_path(delivery, path)
    File.join(
        delivery.e_package.outer_directory_name,
        delivery.e_package.inner_directory_name,
        path
    )
  end

  # ----------------------------------------------------------------------
  # Browse
  # ----------------------------------------------------------------------

  def browse_contents(requester, package_relative_path, params = {})
    #
    # We want to talk to the requester's node.
    #
    if requester == self.sender
      d_or_rd = self
    else
      d_or_rd = self.find_recipient_delivery_for_user(requester)
    end
    #
    # Form path
    #

    share_relative_path = full_path(d_or_rd, package_relative_path)
    path = d_or_rd.storage_share.node_path(share_relative_path)
    #
    # Browse. Returns NodeApi::Response::Browse
    #
    d_or_rd.storage_share.node.browse(path, params)
  end

  # ----------------------------------------------------------------------
  # Delete
  # ----------------------------------------------------------------------

  def delete_files(paths)
    #
    # Form paths
    #
    paths = paths.map {|p|
      self.storage_share.node_path(
          full_path(self, p)
      )
    }

    #
    # Delete. Returns NodeApi::Response::Delete
    #
    self.storage_share.node.api_client.delete(paths)
  end

  # ----------------------------------------------------------------------
  # Rename
  # ----------------------------------------------------------------------

  def rename_files(paths)
    request_hashes = paths.map { |path_hash|
      {
          path:        self.full_package_path(self.storage_share),
          source:      path_hash[:source],
          destination: path_hash[:destination]
      }
    }

    self.storage_share.node.api_client.bulk_rename(request_hashes)
  end

  # ----------------------------------------------------------------------
  # Update / edit
  # ----------------------------------------------------------------------

  # Errors
  class UpdateError < RuntimeError; end

  ##
  # Updates state of delivery and saves to database.
  #
  # This method *should not* be called if you intend to save other information
  # at the same time, like new recipients, and so on.
  #
  # This is intended for internal purposes. Remember that state is read-only
  # from an API perspective.
  def update_state(new_state)
    self.state = new_state
    #
    # Skip bulk of validations.
    #
    self.skip_metadata_validation = true
    self.skip_path_validation = true
    self.skip_source_validation = true
    self.skip_recipient_validation = true
    #
    # Save
    #
    unless self.save
      raise UpdateError,
            "Could not update package's state to <#{new_state}>."
    end
  end

  ##
  # Marks delivery *only* as deleted (not the underlying e_package).
  def mark_as_deleted
    self.update_state(STATE_DELETED)
  end

  ##
  # Deletes old recipients and replaces with new ones. This should
  # be invoked inside a transaction, and it doesn't attempt to
  # save the new recipients to the database (doesn't invoke #save).
  #
  # If either is not present, those recipients are not modified
  # RETURN false if no modification
  def update_recipients(api_request_hash)

    new_public_recipients = api_request_hash[:recipients]
    new_private_recipients = api_request_hash[:private_recipients]

    unless new_public_recipients || new_private_recipients
      #
      # Nothing to do.
      #
      self.skip_recipient_validation = true
      return false
    end

    #
    # With a little luck, we are able to re-use the (awesome) recipient
    # logic. All we need to do is reset @recipient_hashes (and that is
    # probably only necessary for unit tests).
    #
    @recipient_hashes = nil

    if new_public_recipients
      #
      # We have new public recipients; remove all of the old
      # public recipient deliveries.
      #
      self.recipient_deliveries.public.each do |rd|
        rd.delete
      end
      #
      # Set the new pub recips
      #
      self.public_recipients = new_public_recipients
    end

    if new_private_recipients
      #
      # We have new private recipients; remove all of the old
      # private recipient deliveries.
      #
      self.recipient_deliveries.private.each do |rd|
        rd.delete
      end
      #
      # Set the new priv recips
      #
      self.private_recipients = new_private_recipients
    end
    #
    # Make the new recipient deliveries
    #
    make_recipient_deliveries
    return true
  end

  ##
  # Update release policy, release date, and expiration date.
  #
  # Does not invoke #save. (You should invoke this inside a transaction.)
  #
  # If a field is not present, that field is not updated.
  #
  # Note that we *do not* update the state of the delivery. This will
  # be done automatically by the background process.
  def update_release_policy_and_expiration(api_hash,package_modified=true)
    if api_hash[:release_policy]
      self.release_policy = api_hash[:release_policy]
    end
    #
    # These require a little care because they can be null.
    #
    # These are a little messy right now. TODO: clean up.
    #
    if api_hash.key?(:release_date)
      if api_hash[:release_date]
        begin
          api_hash[:release_date] =
              Delivery.parse_date(api_hash[:release_date])
        rescue
          raise UpdateError,
                "Bad release date: <#{api_hash[:release_date]}>"
        end
      end
      self.release_date = api_hash[:release_date]
    elsif (self.release_policy == RELEASE_POLICY_NOW) && package_modified
      self.release_date = Time.now
    end
    if api_hash.key?(:expiration_date)
      if api_hash[:expiration_date]
        begin
          api_hash[:expiration_date] =
              Delivery.parse_date(api_hash[:expiration_date])
        rescue
          raise UpdateError,
                "Bad expiration date: <#{api_hash[:expiration_date]}>"
        end
      end
      self.expiration_date = api_hash[:expiration_date]
    elsif EConfiguration.packages_expire? && package_modified
      after_release = EConfiguration.get_days_until_packages_expire
      logger.warn "Setting expiration date to #{after_release} days after release"
      self.expiration_date = self.release_date ? self.release_date + after_release.days : nil
    end
  end

  ##
  # Updates the delivery's archived flag. This is the *sender's*
  # archived flag. For recipients, you need to find the correct
  # recipient delivery and mark that as archived.
  # RETURN false, this does not modify the package
  def update_archived(api_hash)
    if api_hash.key?(:archived)
      self.archived = api_hash[:archived]
    end
    return false
  end

  ##
  # Updates the delivery's title and note, if they're present.
  # RETURN false if no modification
  def update_title_note(api_hash)
    return false if  !api_hash[:title] and  !api_hash[:note]
    if api_hash[:title]
      self.title = api_hash[:title]
    end
    if api_hash[:note]
      self.note = api_hash[:note]
    end
    return true
  end

  ##
  # Updates the delivery's metadata, if present.
  # RETURN false if no modification
  def update_metadata(api_hash)
    unless api_hash[:metadata]
      self.skip_metadata_validation = true
      return false
    end
    #
    # Sanitize new metadata.
    #
    new_metadata = api_hash[:metadata]
    new_metadata =
        MetadataProfile.sanitize_metadata_array_using_active_profile(
            new_metadata,
            self.dropbox)
    #
    # Get old metadata
    #
    old_metadata = ActiveSupport::JSON.decode(self.metadata_json)
    #
    # Set system fields in new metadata
    #
    #add_system_metadata_to_hash(new_metadata)
    #
    # Overwrite created utc with old
    #
    new_metadata['_created_utc'] = old_metadata['_created_utc']
    #
    # Set fields
    #
    self.metadata = new_metadata
    self.metadata_json = new_metadata.to_json
    return true
  end

  ##
  # Updates download limit, if present.
  # RETURN false if no modification
  def update_download_limit(api_hash)
    if api_hash.key?(:download_limit)
      self.download_limit = api_hash[:download_limit]
      return true
    else
      return false
    end
  end

  ##
  # Controls updates of archived flag for deliveries in a 'transient held to released' state
  def update_transient_routing_from_hash(requester,api_request_hash)
    # you can only update the archived flag IF THE RECIPIENT has
    # correctly received the package
    unless api_request_hash.key?(:archived)
      raise UpdateError, "Package is in #{self.state} state. Only 'archived' flag can be modified at this time."
    end
    # case 1 requester == sender
    # case 2 requester == recipient
    if requester == self.sender
      raise UpdateError , "Package cannot be modified in #{self.state} state" if self.upload_status!= ESession::STATUS_COMPLETE
    else
      rd = UserDelivery.find_by_recipient_id_and_delivery_id(requester.id,self.id)
      raise UpdateError , "Package cannot be modified in #{self.state} state"  if  (  rd.nil? or  rd.upload_status!= ESession::STATUS_COMPLETE)
    end
    if api_request_hash[:mailbox].present?
      self.update_recipient_archived(requester, api_request_hash) if api_request_hash[:mailbox] == MAILBOX_INBOX
      self.update_non_released_from_hash(requester, api_request_hash) if api_request_hash[:mailbox] == MAILBOX_OUTBOX
    else
      # no mailbox passed -> do both (sender and recipient archived)
      self.update_non_released_from_hash(requester, api_request_hash)
      self.update_recipient_archived(requester, api_request_hash)
    end
  end

  ##
  # Updates delivery assuming it is not in released state.
  #
  # Everything is done in a transaction, so it is all or nothing. Raises
  # an UpdateError if update fails.
  def update_non_released_from_hash(requester, api_request_hash)
    #
    # If requester is not sender, silently ignore.
    #
    unless requester == self.sender
      return
    end
    #
    # We have to reset these
    #
    self.invalid_recipient_names = []
    self.warnings = {}

    updated = false
    modified = false
    Delivery.transaction do
      #
      # Update title and note
      #
      modified = self.update_title_note(api_request_hash) || modified
      #
      # Update recipients
      #
      modified = self.update_recipients(api_request_hash) || modified
      #
      # Update archived
      #
      modified = self.update_archived(api_request_hash) || modified
      #
      # Update download limit
      #
      modified = self.update_download_limit(api_request_hash) || modified
      #
      # Update metadata
      #
      # NOTE: We do metadata last because it is affected by the active
      # profile, and we set some system metadata in here. (These are
      # things we may be in the process of updating, so they need to
      # be consistent.)
      #
      modified = self.update_metadata(api_request_hash) || modified
      #
      # Update release policy, release date, and expiration date.
      #
      self.update_release_policy_and_expiration(api_request_hash,modified)
      #
      # We never update sources or paths. (No content editing here.)
      #
      self.skip_source_validation = true
      self.skip_path_validation = true
      #
      # Attempt save.
      #
      self.updated_at = Time.now
      self.e_package.updated_at = Time.now
      if self.save
        updated = true
      else
        raise ActiveRecord::Rollback
      end
    end
    unless updated
      raise UpdateError,
            self.formatted_error_messages
    end
    self.touch
    self.e_package.touch
  end

  ##
  # Updates a recipient's recipient delivery to be archived, if present.
  #
  # Unlike other update_* methods, this **does** update the db immediately.
  def update_recipient_archived(recipient, api_hash)
    unless api_hash.key?(:archived)
      return # nothing to do
    end
    #
    # Find first available recipient delivery.
    #
    rd = self.find_recipient_delivery_for_user(recipient)
    #
    # If not found, we're in trouble.
    #
    unless rd
      raise UpdateError,
            "Failed to find recipient information for #{recipient.name}."
    end
    #
    # Update archived.
    #
    rd.update_attribute(:archived, api_hash[:archived])
  end

  ##
  # Does updates assuming delivery is released.
  def update_released_from_hash(requester, api_request_hash)
    #
    # If requester is not the sender / creator, they can only mark
    # the (recipient) delivery as archived.
    #
    unless requester == self.sender
      #
      # Do archived update, if applicable.
      #
      self.update_recipient_archived(requester, api_request_hash)
      return
    end
    #
    # For now, we just do the same thing as if the delivery is not
    # released.
    #
    # TODO: Need to route to new recipients, make inaccessible to old
    # ones.
    #
    # TODO: Send out notifications.
    #
    self.update_non_released_from_hash(requester, api_request_hash)
  end

  ##
  # Updates delivery using values in hash.
  #
  # Raises UpdateError if there is a problem updating the delivery, or if
  # the delivery is in a state in which it cannot be updated (e.g. if it
  # is deleted).
  def update_from_hash(requester, api_request_hash)
    #
    # We completely separate the two cases since the
    # logic is way different for certain updates - like updating
    # recipients.
    #
    case self.state
      when STATE_RELEASED
        self.update_released_from_hash(requester, api_request_hash)
      when STATE_HELD, STATE_EXPIRED
        self.update_non_released_from_hash(requester, api_request_hash)
      when TRANSIENT_STATE_HELD_TO_RELEASED_ROUTING_PACKAGES
        self.update_transient_routing_from_hash(requester, api_request_hash)
      else
        raise UpdateError,
              "Package in state #{self.state}. You cannot update it at this time."
    end
  end

  # ----------------------------------------------------------------------
  # Data
  # ----------------------------------------------------------------------

  ##
  # Returns (non-system) metadata fields as an array:
  #
  #   [
  #      { :name => 'field1', :value => value1 },
  #      { :name => 'field2', :value => value2 }
  #   ]
  #
  def user_metadata_to_array
    out = []
    user_metadata_hash.each do |key, val|
      out << { :name => key.to_s, :value => val }
    end
    out
  end

  def self.date_to_iso_string_format(d)
    #
    # Convert to DateTime. Time has weird bug (?) when you try to use
    # %3N.
    #
    if d.is_a?(Time)
      d = d.to_datetime
    end
    d.strftime("%FT%T.%3N%z")
  end

  def notifiable_to_array(field)
    if field.is_a?(Array)
      field
    elsif field.is_a?(String)
      field.split(',')
    else
      []
    end
  end

  ##
  # Returns the hash expected for the Faspex Overlay REST API.
  #
  # *IMPORTANT*
  #
  # This should not be used in the controllers; use the 'view' methods that
  # follow; they are 'contextualized' based on the requesting user
  # and so on (so that for example sensitive doesn't go out to recipients).
  def to_faspex_overlay_api_hash
    #
    # Doing it this way so we can conditionally set certain fields (i.e., they
    # won't show up as being set to null. Requested by Guillaume.).
    #
    # Not my fav way of doing this, but it works ...
    #
    out = {}
    out[:id] = id.to_s
    out[:title] = title
    if note
      out[:note] = note
    end

    all_recipients = recipient_deliveries
    rs = all_recipients.select{ |r| !r.private }
    prs = all_recipients.select{ |r| r.private }
    if rs.length > 0
      out[:recipients] = rs
    end
    if prs.length > 0
      out[:private_recipients] = prs
    end
    if user_metadata_to_array.length > 0
      out[:metadata] = user_metadata_to_array
    end
    out[:release_policy] = release_policy
    if release_date
      out[:release_date] = Delivery.date_to_iso_string_format(release_date)
    end
    if expiration_date
      out[:expiration_date] = Delivery.date_to_iso_string_format(expiration_date)
    end
    out[:sender] = sender_name
    if download_limit
      out[:download_limit] = download_limit
    end
    if state.include? ('transient_') # don't show client transient states
      out[:state] = state.sub(/transient_/, '').sub(/_.*$/, '')
    else
      out[:state] = state
    end

    # TODO - Make this less of a hack. 'Forward' is not a valid state
    out[:state] = 'held' if out[:state] == 'forward'

    out[:archived] = archived?
    out[:ear_enabled] = ear_enabled
    if notify_on_upload
      out[:notify_on_upload] = notify_on_upload
    end
    if notify_on_download
      out[:notify_on_download] = notify_on_download
    end

    out[:notifiable_on_upload] = notifiable_to_array(notifiable_on_upload)
    out[:notifiable_on_download] = notifiable_to_array(notifiable_on_download)

#    downloaders = []
#    e_package.completed_downloads do |downloader|
#      downloaders.push downloader[:initiator_name]
#    end

#    active_downloaders = []
#    e_package.active_downloads do |downloader|
#      downloaders.push downloader[:initiator_name]
#    end

    sessions = e_package.e_sessions
    active_downloads = sessions.select{ |s| [ESession::STATUS_TRANSFERRING, ESession::STATUS_TO_BE_VALIDATED, ESession::STATUS_VALIDATING].include?(s.status) && s.direction == ESession::DIRECTION_DOWNLOAD }
    completed_downloads = sessions.select{ |s| s.status == ESession::STATUS_COMPLETE && s.direction == ESession::DIRECTION_DOWNLOAD }

    ### MK changes START
    # -- Add transfer_request_uuid from Cookies as a token_uuid to returned hash to allow Speedracer to reconcile download sessions based on the transfer_request_uuid
    if active_downloads.length > 0
      active_downloads.each { |s| s[:transfer_request_uuid] = s.transfer_request.uuid }
    end
    if completed_downloads.length > 0
      completed_downloads.each { |s| s[:transfer_request_uuid] = s.transfer_request.uuid }
    end
    #    logger.error("MK5: in Delivery.to_faspex_overlay_api_hash: survived completed_downloads: #{completed_downloads.inspect}")
    ### MK changes END

    out[:active_downloads] = active_downloads.length
    out[:active_downloaders] = active_downloads
    out[:download_count] = completed_downloads.length
    out[:downloaders] = completed_downloads

    out[:total_bytes] = e_package.e_package_total_size_bytes
    out[:total_files] = e_package.e_package_total_files
    out[:recalculation_needed] = e_package.recalculation_needed?
    out[:recalculation_in_progress] = e_package.recalculation_in_progress?
    out[:creation_date] = Delivery.date_to_iso_string_format(created_at)
    out[:last_modified] = Delivery.date_to_iso_string_format(e_package.updated_at)
    out[:package_uuid] = e_package.e_package_uuid
#
# Return hash
#
    out
  end

  # Figure out whether to access delivery or recipient delivery
  def delivery_or_recipient_delivery!(user, api_request_hash)
    rd = self.find_recipient_delivery_for_user(user)
    if api_request_hash[:mailbox] == Delivery::MAILBOX_OUTBOX
      # Sent version
      # Throw an error if a non-admin is trying to access another user's package
      raise Api::BadRequestError, "You may not modify a package you did not send" unless user.admin? || self.sender_id == user.id

      self
    elsif api_request_hash[:mailbox] == Delivery::MAILBOX_INBOX || user.id != self.sender_id
      # Received version
      # Throw an error if specifically looking for invalid recipient delivery
      raise Api::BadRequestError, "No such recipient delivery for package ID #{api_request_hash[:package_id]} for user #{user.name}" unless rd.present?

      rd
    else
      # Sent by current user - May return either version, default to inbox if sent to self
      rd || self
    end
  end

  def view_for_faspex_overlay_user(user, api_request_hash)
    #
    # Get (pretty much) full hash
    #
    api_hash = self.to_faspex_overlay_api_hash
    #
    # If user is not sender, delete some sensitive fields.
    #
    if user != self.sender
      api_hash.delete(:private_recipients)  unless user.super_admin?
      api_hash.delete(:ear_passphrase)
      api_hash.delete(:notify_on_upload)
      api_hash.delete(:notify_on_download)
      api_hash.delete(:notifiable_on_upload)
      api_hash.delete(:notifiable_on_download)
      # remove the downloader as the UI does . only the sender see these info
      api_hash.delete(:active_downloads) unless user.super_admin?
      api_hash.delete(:active_downloaders)  unless user.super_admin?
      api_hash.delete(:download_count)  unless user.super_admin?
      api_hash.delete(:downloaders)  unless user.super_admin?
    end
    unless user.super_admin?
      # DO NOT SEND THE PASS CODE TO ANYONE -- either to the sender unless is admin
      api_hash[:recipients].each {|r| r.send_to_email_passcode=nil}  if api_hash[:recipients].present?
    end
    #
    # archived; user-dependent
    #

    d_or_rd = delivery_or_recipient_delivery!(user, api_request_hash)

    api_hash[:archived] = d_or_rd.is_a?(Delivery) ? d_or_rd.archived? : d_or_rd.delivery.archived?
    api_hash[:mailbox] = d_or_rd.is_a?(Delivery) ? MAILBOX_OUTBOX : MAILBOX_INBOX

    #
    # Return hash
    #
    api_hash
  end

  def view_for_any_user
    #
    # Get (pretty much) full hash
    #
    api_hash = self.to_faspex_overlay_api_hash
    #
    # If user is not sender, delete some sensitive fields.
    #

    api_hash.delete(:private_recipients)
    api_hash.delete(:ear_passphrase)
    api_hash.delete(:notify_on_upload)
    api_hash.delete(:notify_on_download)
    api_hash.delete(:notifiable_on_upload)
    api_hash.delete(:notifiable_on_download)

    api_hash[:archived] = self.archived?
    #
    # Return hash
    #
    api_hash
  end
  # ----------------------------------------------------------------------
  # Lookup
  # ----------------------------------------------------------------------

  # Errors
  class LookupError < RuntimeError; end

  def find_recipient_delivery_for_user(user)
    #
    # Look for user delivery first.
    #
    ud = UserDelivery.find_for_user_and_delivery(user, self)
    return ud if ud.present?

    #
    # Look for (first) workgroup delivery next. Will be nil if none found.
    #
    wd = WorkgroupDelivery.find_first_for_user_and_delivery(user, self)
    return wd if wd.present?

    if user.admin?
      self
    end
  end

  def self.find_and_check_perms(delivery_id, requester)
    #
    # case 1: requester is sender
    #
    delivery = Delivery.find_by_id(delivery_id)
    unless delivery
      #
      # Doesn't exist at all.
      #
      return nil
    end
    #
    # If delivery is deleted, pretend it doesn't even exist.
    #
    if delivery.state == STATE_DELETED
      return nil
    end
    #
    # Allow sender access
    #
    if delivery.sender == requester
      return delivery
    end
    #
    # case 2: requester is recipient
    #
    # Don't let recipients see held or expired deliveries.
    #
    unless delivery.state == STATE_RELEASED
      #subcase-- the package can be a relay and a relay in transit, we need to understand
      #if the package is available at the share for this requester
      rd = UserDelivery.find_by_recipient_id_and_delivery_id(requester.id,delivery_id)
      return delivery if  ( !rd.nil? && rd.upload_status == ESession::STATUS_COMPLETE)
      return nil
    end
    #
    # Now ensure they are a recipient
    #
    if delivery.find_recipient_delivery_for_user(requester)
      return delivery
    else
      #
      # Requester is neither sender nor recipient.
      #
      return nil
    end
  end

  def self.pagination_limit_sql(hash)
    start_index = (hash[:startIndex] || 0).to_i
    count = (hash[:count] || DEFAULT_RECORDS_PER_PAGE).to_i

    "#{start_index}, #{count}"
  end

  def self.recipient_ids(requester, hash, all_users_allowed)
    if all_users_allowed && hash[:recipients].present?
      recipients = hash[:recipients].map{ |r|
        user = User.find_by_name(r)
        raise Api::BadRequestError, "Recipient not found: #{r}" unless user.present?
        user
      }.compact
      recipients.map(&:id)
    elsif requester.present?
      [requester.id]
    else
      nil
    end
  end

  def self.sender_ids(requester, hash, all_users_allowed)
    if all_users_allowed && hash[:senders].present?
      senders = hash[:senders].map{ |s|
        user = User.find_by_name(s)
        raise Api::BadRequestError, "Sender not found: #{s}" unless user.present?
        user
      }.compact
      senders.map(&:id)
    elsif requester.present?
      [requester.id]
    else
      nil
    end
  end

  def self.state_to_sql(hash, outbox)
    if hash.has_key?(:state)
      state = hash[:state]
      if state.is_a?(Array)
        state = state.map{|s| "#{sanitize s}"}
        "(#{state.join(',')})"
      else
        "(#{sanitize state})"
      end
    else
      # State not specified
      if outbox
        "('#{STATE_HELD}', '#{STATE_RELEASED}')"
      else
        "('#{STATE_RELEASED}')"
      end
    end
  end

  def self.search_deliveries(requester, api_request_hash, all_users_allowed=false)
    api_request_hash = api_request_hash.symbolize_keys

    conditions = []
    vars = []

    state_list = state_to_sql(api_request_hash, api_request_hash[:mailbox] == MAILBOX_OUTBOX)
    state_str = state_list.present? ? " AND state IN #{state_list}" : ""

    # Outbox
    if api_request_hash[:mailbox] != MAILBOX_INBOX
      senders = sender_ids(requester, api_request_hash, all_users_allowed)
      if senders.present?
        senders_sql = "(#{senders.join(',')})"
        conditions << "(sender_id in #{senders_sql} #{state_str})"
      end
    end

    # Inbox
    if api_request_hash[:mailbox] != MAILBOX_OUTBOX
        recipients = recipient_ids(requester, api_request_hash, all_users_allowed)
      if recipients.present?
        # Normal recipients
        recipient_sql = "(#{recipients.join(',')})"
        sql = "(recipient_deliveries.recipient_type = ? AND recipient_id in #{recipient_sql} #{state_str})"
        vars += [User.name]

        workgroup_ids = EMembership.find_all_by_user_id_and_can_read(recipients, true).map(&:e_workgroup_id)
        # Released workgroup deliveries
        if workgroup_ids.any?
          ids = "(#{workgroup_ids.join(',')})"
          sql += " OR (recipient_deliveries.recipient_type = ? AND recipient_id IN #{ids} AND state = ?)"
          vars += [EWorkgroup.name, STATE_RELEASED]
        end

        conditions << "( #{sql} ) "
      end
    end

    conditions = conditions.any? ? ["( #{conditions.join(' OR ')} )"] : []

    if api_request_hash.has_key?(:archived)
      conditions << "recipient_deliveries.archived = ?"
      vars << (['true', true].include?(api_request_hash[:archived]) ? 1 : 0)
    end

    if api_request_hash[:title].present?
      conditions << "deliveries.title LIKE ?"
      vars << "%#{ api_request_hash[:title]}%"
    end

    joins = [:recipient_deliveries, :e_package]
    order = api_request_hash[:order] == 'a' ? 'ASC' : 'DESC'
    sort_by = 'created_at'
    sort_table = "deliveries"
    result_order = "`#{sort_table.shellescape}`.#{sort_by} #{order}"

    if api_request_hash[:sort].present?
      result_order = self.parse_sorting(api_request_hash[:sort], api_request_hash[:order])
    end

    if api_request_hash[:changedSinceDate].present?
      changed_date = Time.parse(api_request_hash[:changedSinceDate]) rescue ""
      if changed_date.present?
        conditions << "(deliveries.updated_at >= ? OR recipient_deliveries.updated_at >= ? OR e_packages.updated_at >= ?)"
        vars += [changed_date, changed_date, changed_date]
      end
    end

    Delivery.all(
        include: joins,
        conditions: [conditions.join(' AND '), *vars],
        order: result_order,
        limit: pagination_limit_sql(api_request_hash)
    )
  end

  def self.parse_sorting(cols, order)
    columns = cols.split(',').reject(&:blank?)
    orders = order.present? ? order.split(',').reject(&:blank?).map(&:downcase) : []
    order_by = []

    columns.each_with_index do |col, i|
      ord = orders[i].present? && ['a', 'd'].include?(orders[i]) ? orders[i] : "DESC"
      if ALLOWED_SORT_FIELDS.include?(col)
        order_by << "`deliveries`.#{col} #{ord}"
      elsif ALLOWED_PACKAGE_SORT_FIELDS.include?(col)
        order_by << "`e_packages`.#{col} #{ord}"
      end
    end

    order_by.join(', ')
  end

  def self.faspex_overlay_api_search(requester, api_request_hash, all_users_allowed=false)
    if api_request_hash.has_key? 'state'
      api_request_hash['state'] += ['transient_held_to_released_start', 'transient_held_to_released_routing_packages', 'transient_held_to_released_finish', 'transient_held_to_expired', 'transient_held_to_deleted']  if api_request_hash['state'].include?('held')
      api_request_hash['state'] += ['transient_released_to_expired', 'transient_released_to_deleted']  if api_request_hash['state'].include?('released')
      api_request_hash['state'] += ['transient_expired_to_deleted']  if api_request_hash['state'].include?('expired')
    end
    self.search_deliveries(requester, api_request_hash, all_users_allowed)
  end

  # ----------------------------------------------------------------------
  # Transfer specs
  # ----------------------------------------------------------------------

  # Errors
  class TransferSpecError < RuntimeError; end

  def upload_spec_for_user(requester, request_ip, request_hostname, paths=['/'])
    #
    # Ensure requester is sender
    #
    if self.sender != requester
      raise TransferSpecError,
            'You, as a recipient, can\'t modify the package\'s content.'
    end
    #
    # Get the transfer spec. The requester's assigned share will match
    # the delivery's storage share.
    #
    host = self.remote_host(request_ip, request_hostname)
    tspec = self.get_transfer_spec(ESession::DIRECTION_UPLOAD,
                                   paths,
                                   requester,
                                   token: true,
                                   for_fasp_url: true,
                                   host: host)
    tspec.content_protection = "encrypt" if ear_encrypted?
    #
    # Return spec
    #
    tspec
  end

  def download_spec_for_user(requester, request_ip, request_hostname, override_shares = [], paths=['/'])
    #
    # If requester is sender, we use the delivery itself. If requester is
    # recipient, we use their recipient delivery.
    #
    if requester == self.sender
      d_or_rd = self
    else
      d_or_rd = self.find_recipient_delivery_for_user(requester)
    end
    #
    # Get the transfer spec. This will use the requester's assigned share.
    #
    self.e_package.storage_share = Share.default_destination if nil == self.storage_share
    host = d_or_rd.remote_host(request_ip, request_hostname)
    tspec = d_or_rd.get_transfer_spec(ESession::DIRECTION_DOWNLOAD,
                                      paths,
                                      requester,
                                      token: true,
                                      for_fasp_url: true,
                                      host: host,
                                      share_id_order: override_shares,
                                      request_ip: request_ip,
                                      request_hostname: request_hostname)

    tspec.content_protection = "decrypt" if ear_encrypted?
    #
    # Return spec
    #
    tspec
  end

  # ----------------------------------------------------------------------
  # Progress
  # ----------------------------------------------------------------------

  def upload_stats_for_user(requester)
    #
    # If requester is not sender, just return empty array.
    #
    if self.sender != requester
      return []
    end
    #
    # Get all upload stats
    #
    sessions = delivery.e_package.e_sessions.all(:conditions => [%q{
direction = ? OR direction = ?
},
                                                                 ESession::DIRECTION_UPLOAD,
                                                                 ESession::DIRECTION_CREATE])
    #
    # Return array of stats
    #
    sessions.map(&:to_faspex_overlay_api_stats)
  end

  def download_stats_for_user(requester)

    if self.sender == requester
      #
      # Requester is sender - show all.
      #
      delivery.e_package.e_sessions.downloads.map(&:to_faspex_overlay_api_stats)
    else
      #
      # If requester is recipient, show theirs only.
      #
      sessions = delivery.e_package.e_sessions.downloads.all(
          :conditions => [
              'initiator_user_id = ?',
              requester.id
          ])
      #
      # Return array of stats
      #
      sessions.map(&:to_faspex_overlay_api_stats)
    end
  end

  def transfer_stats_for_user(requester, api_request_hash)
    case api_request_hash[:direction]
      when TRANSFER_DIRECTION_SEND
        return self.upload_stats_for_user(requester)
      when TRANSFER_DIRECTION_RECEIVE
        return self.download_stats_for_user(requester)
      else
        return (self.upload_stats_for_user(requester) +
            self.download_stats_for_user(requester))
    end
  end

  # ----------------------------------------------------------------------
  # Remote Transfer
  # ----------------------------------------------------------------------

  # Errors
  class RemoteTransferBadParamsError < RuntimeError; end
  class RemoteTransferServerError < RuntimeError; end

  def initiate_remote_transfer_for_user_and_params(requester, api_request_hash)
    #
    # If requester is not sender, reject.
    #
    unless requester == self.sender
      raise RemoteTransferBadParamsError,
            'You are not the creator of the package, so you cannot edit it.'
    end
    #
    # Must be direction = send (this might be silly ...)
    #
    unless api_request_hash[:direction] == TRANSFER_DIRECTION_SEND
      raise RemoteTransferBadParamsError,
            "Bad direction: <#{api_request_hash[:direction]}>."
    end
    #
    # Paths must be present and be an array of unique items, with at least 1
    #
    share_paths = api_request_hash[:paths]
    error_message = Delivery.do_check_paths(share_paths)
    if error_message
      raise RemoteTransferBadParamsError,
            "Bad paths: <#{share_paths}>. Error: #{error_message}"
    end
    #
    # Source share must be present and valid, and sender must have
    # rights to do transfer.
    #
    share = SourceShare.find_by_id(api_request_hash[:share_id])
    unless share
      raise RemoteTransferBadParamsError,
            "Bad source share id: <#{api_request_hash[:share_id]}>."
    end
    unless user_can_transfer_from_share?(requester, share)
      raise RemoteTransferBadParamsError,
            "Requester #{requester.name} is not authorized to send from " +
                "share #{share.name}."
    end

    # Convert share relative paths to node paths
    share_paths = share.node_paths(share_paths)

    #
    # Now we're ready for action ...
    #
    # Create transfer job for node-to-node transfer. The source paths
    # will be transferred to the
    #
    transfer_job = create_upload_job(share, share_paths, false)
    #
    # Attempt to fire it off
    #
    begin
      transfer_job.perform
    rescue
      raise RemoteTransferServerError,
            $!.message
    end
    #
    # Create stub session so we can return stats immediately to
    # requester.
    #
    stub_session = ESession.make_stubbed_e_session_for_node_to_node_transfer(
        requester, self, transfer_job)
    #
    # Return stubbed session stats
    #
    stub_session.to_faspex_overlay_api_stats
  end

  # ---------------------------------------------------------------------- #
  # ---------------------------------------------------------------------- #
  #                 STATE MACHINE / BACKGROUND LOGIC                       #
  #                                                                        #
  # This is all executed by a background thread / process (configured in   #
  # lib/daemons/background/background_worker.rb).                          #
  # ---------------------------------------------------------------------- #
  # ---------------------------------------------------------------------- #

  # ----------------------------------------------------------------------
  # HELD => EXPIRED
  # ----------------------------------------------------------------------

  def do_held_to_expired
    #
    # Nothing special to do since the delivery hasn't been made available to
    # recipients. Just mark the delivery as expired.
    #
    self.update_state(Delivery::STATE_EXPIRED)
  end

  def self.get_held_to_expired_deliveries
    deliveries = []

    Delivery.transaction do
      #
      # Look up all deliveries that are held and
      # expiration date <= now.
      #
      deliveries = Delivery.all(
        :conditions => [
          'state = ? AND expiration_date <= ?',
          Delivery::STATE_HELD,
          Time.now
        ]
      )

      #
      # Put them in transient state so no one else will interfere or
      # duplicate our work.
      #
      deliveries.each do |d|
        d.update_state(Delivery::TRANSIENT_STATE_HELD_TO_EXPIRED)
      end
    end

    deliveries
  end

  def self.handle_deleted_deliveries
    deliveries = Delivery.all(
      :conditions => [
        'state <> ? AND state <> ? AND e_packages.e_package_files_deleted_by_admin_time IS NOT NULL',
        STATE_DELETED,
        STATE_EXPIRED
      ],
      :joins => :e_package
    )

    deliveries.each do |d|
      Delivery.find(d.id).update_state(STATE_DELETED)
    end
  end

  def self.handle_held_to_expired_deliveries

    deliveries = self.get_held_to_expired_deliveries
    #
    # Process held => expired for each.
    #
    deliveries.each do |d|
      d.do_held_to_expired
    end
  end

  # ----------------------------------------------------------------------
  # RELEASED => EXPIRED
  # ----------------------------------------------------------------------

  def do_released_to_expired
    #
    # TODO: Error handling logic?
    #
    # Delete package from those recipients' shares that differ from
    # the creator's share.
    #
    self.delete_from_recipient_shares
    #
    # Mark as expired.
    #
    delivery.update_state(Delivery::STATE_EXPIRED)
  end

  def self.get_deliveries_for_released_to_expired
    deliveries = []

    Delivery.transaction do
      #
      # Look up all deliveries that are released and
      # expiration date <= now.
      #
      deliveries = Delivery.all(
        :conditions => [
          'state = ? AND expiration_date <= ?',
          Delivery::STATE_RELEASED,
          Time.now
        ]
      )
      #
      # Put them in transient state so no one interferes.
      #
      deliveries.each do |d|
        d.update_state(Delivery::TRANSIENT_STATE_RELEASED_TO_EXPIRED)
      end
    end

    deliveries
  end

  def self.handle_released_to_expired_deliveries

    deliveries = self.get_deliveries_for_released_to_expired
    #
    # Process live => expired for each.
    #
    deliveries.each do |d|
      d.do_released_to_expired
    end

  end

  # ------------------------------------------------------------------
  # HELD => RELEASED
  #
  # There are two steps - (1) kicking off the package routing, and
  # (2) detecting that it's complete and marking the delivery as released
  #
  # ------------------------------------------------------------------

  def do_held_to_released_step_1_except_state
    #
    # NOTE: This used to be in EPackage#handle_upload_end. All of the logic
    # was moved into the body of this method to avoid confusion. Lines that
    # are direct copies of that code are noted with "faspex 3.0".
    #
    # ----------------------------------------------------------------------
    #
    # Get the underlying e_package.
    #
    e_package = self.e_package
    #
    # Force EAR if creator used some other means to encrypt files. (faspex 3.0)
    #
    e_package.check_ear_encrypted
    #
    # Send upload notification to creator and those to be notified, unless
    # this is an inter-faspex relay. (faspex 3.0)
    #
    # TODO: This should probably change to "delivery go live notification"?.
    #
    e_package.deliver_upload_result(self) unless self.relay?

    if invitation = self.invitation
      #
      # Handle external submissions. (faspex 3.0)
      #
      e_package.handle_external_submission_end(invitation)
    elsif e_package.dropbox_package?
      #
      # Handle dropbox submissions. (faspex 3.0)
      #
      e_package.handle_dropbox_upload_end(self)
    end
    #
    # Fire off any post process scripts (faspex 3.0)
    #
    PostProcess.match_and_execute(self)
    #
    # Generate a sequence id (faspex 3.0). (As far as I know, this is only
    # needed so that deliveries can be paginated in the xml atom feed ...)
    #
    self.generate_sequence_id!
    #
    # Remove user/workgroup deliveries that are no longer valid (ie deleted/deactivated etc)
    self.remove_invalid_recipient_deliveries
    # Do regular package routing for users who download from shares
    # different from the e_package's storage share. (faspex 3.0).
    #
    # NOTE: This innocent little method is what takes care of package
    # routing.
    #
    self.relay_to_inboxes unless self.relay?
    #
    # Notify those users who receive packages on the delivery's storage
    # share. (faspex 3.0)
    #
    self.send_received_notifications(self.storage_share)
    #self.destination_shares.map{ |d| self.send_received_notifications(d) }
    self.send_receipt_cc_notifications
    #
    # Set up inter-faspex relays (will be handled by a background thread).
    # (faspex 3.0)
    #
    e_package.create_relays_for(self) unless self.relay?
    #
    # Do workgroup relays (this is not the same thing as inter-faspex
    # relays, nor is it the same thing as package routing). (faspex 3.0)
    #
    e_package.relay_to_forwarding_destinations(self)
  end

  def do_held_to_released_step_1

    self.do_held_to_released_step_1_except_state
    #
    # -----------------------------------------------------------
    #
    # At this point, we've kicked off the routing process for all the
    # recipients.
    #
    # Mark as held-to-live with package routing in progress.
    #
    self.update_state(
        Delivery::TRANSIENT_STATE_HELD_TO_RELEASED_ROUTING_PACKAGES)
  end

  def self.get_deliveries_for_held_to_released_step_1

    deliveries = []

    Delivery.transaction do
      #
      # This beefy query does the following:
      #
      # Look up all deliveries that are held,
      # release_policy = release_now or release_on_date with date <= now,
      # and whose e_package has at least one completed upload session.
      #
      # As HTTP Gateway uses a separate upload session for each file,
      # also check that there are no upload sessions for this delivery that
      # are not completed.
      #
      # Note that we don't check if it's expired; we did that already
      # in handle_held_to_expired. (Yes, the package could expire in between
      # that method and this one, but we'll just expire right after we
      # go into a released state.)
      #

      deliveries = Delivery.find_by_sql([ %q{
SELECT
        d.*
FROM
        deliveries AS d
        INNER JOIN
                e_packages AS ep
                ON ep.id = d.e_package_id
WHERE
        (
                d.state = ?
                AND
                (
                        d.release_policy = ?
                        OR
                        (
                                d.release_policy = ?
                                AND d.release_date <= ?
                        )
                )
                AND
                (
                        (
                                SELECT
                                        COUNT(*)
                                        FROM
                                                e_sessions AS es
                                        WHERE
                                                es.e_package_id = ep.id
                                        AND
                                        (
                                                es.direction = ?
                                                OR
                                                es.direction = ?
                                        )
                                        AND
                                        es.status = ?
                        ) > 0
                )
                AND
                (
                        NOT ep.http_gateway_transfer
                        OR
                        (
                                SELECT
                                        COUNT(*)
                                FROM
                                        e_sessions AS es
                                WHERE
                                        es.e_package_id = ep.id
                                        AND
                                        (
                                                es.direction = ?
                                                OR
                                                es.direction = ?
                                        )
                                        AND
                                        es.status <> ?
                        ) = 0
                )
        )
        }, Delivery::STATE_HELD, Delivery::RELEASE_POLICY_NOW, Delivery::RELEASE_POLICY_ON_DATE, Time.now, ESession::DIRECTION_UPLOAD, ESession::DIRECTION_CREATE, ESession::STATUS_COMPLETE, ESession::DIRECTION_UPLOAD, ESession::DIRECTION_CREATE, ESession::STATUS_COMPLETE])
      #
      # Put them in transient state.
      #
      deliveries.each do |d|
          d.update_state(Delivery::TRANSIENT_STATE_HELD_TO_RELEASED_START)
      end
    end

    deliveries
  end

  def self.handle_held_to_released_step_1

    deliveries = self.get_deliveries_for_held_to_released_step_1
    #
    # Process held => routing packages for each.
    #
    deliveries.each do |d|
      if d.sender.present? && d.sender.account_activated
        d.do_held_to_released_step_1
      else
        d.update_attributes({:release_policy => Delivery::RELEASE_POLICY_NEVER,
                             :release_date => nil,
                             :skip_metadata_validation => true,
                             :skip_path_validation => true,
                             :skip_source_validation => true,
                             :skip_source_validation => true,
                             :skip_recipient_validation => true})
      end
    end
  end

  # Select deliveries where:
  # - distinct destination shares > 1 (i.e. there is a relay)
  # - there is not a completed session for each share (i.e. not all relays are completed and/or one or more has failed)
  def self.with_incomplete_or_failed_relays(page, per_page, search = nil, order='id DESC')
    conditions = %q(
          ((@n_shares := (SELECT COUNT(DISTINCT share_id) FROM recipient_deliveries rds WHERE rds.delivery_id=deliveries.id AND rds.share_id <> e_packages.storage_share_id)) > 0)
            AND
        @n_shares > (SELECT COUNT(DISTINCT share_id) FROM e_sessions es WHERE es.delivery_id=deliveries.id AND es.status=:completed AND es.share_id <> e_packages.storage_share_id)
            AND
        e_package_status=:valid_package_status
    )

    if search.present?
      conditions = %q(
          (sender_name LIKE :search OR
           title       LIKE :search OR
           note        LIKE :search OR
           (SELECT COUNT(*) FROM recipient_deliveries rds WHERE rds.recipient_name LIKE :search)) > 0 AND
      ) + conditions
    end

    # Need to select extra fields to satisfy after_initialize methods
    delivery_ids = Delivery.find( :all,
      joins: :e_package,
      select: 'deliveries.id, deliveries.sender_id, dbox_extusr_sender_pkg_passcode',
      conditions: [conditions, { search: "%#{search}%", completed: ESession::STATUS_COMPLETE, valid_package_status: EPackage::STATUS_VALID } ]
    )

    result_conditions = delivery_ids.any? ? "deliveries.id in (#{delivery_ids.map(&:id).join(', ')})" : 'false'

    paginate(
      page: page,
      per_page: per_page,
      include: [{e_package: {deliveries: :recipient_deliveries}}, { recipient_deliveries: {share: :node}  }],
      conditions: result_conditions,
      order: order
    )
  end

  # Transfer in progress
  def in_progress_relay_shares
    self.all_recipient_shares.select{ |share| transferring_relay_session_for(share).present? && !relay_succeeded?(share) }.uniq
  end

  # Not in progress or completed, but haven't given up
  def incomplete_relay_shares
    self.all_recipient_shares.select{ |share| !relay_succeeded?(share) }.uniq - self.in_progress_relay_shares - self.failed_relay_shares
  end

  # Gave up
  def failed_relay_shares
    self.all_recipient_shares.select{ |share| relay_failed?(share) }.uniq
  end

  # Succeeded
  def completed_relay_shares
    self.all_recipient_shares.select{ |share| relay_succeeded?(share) }.uniq
  end

  def relay_failed?(share)
    return false if relay_succeeded?(share)

    upload_requests_since_last_connection(share).length > annealed_timeout_periods.length
  end

  def relay_succeeded?(share)
    if share
      upload_requests_for_share(share).any?{ |request| request.status == ESession::STATUS_COMPLETE }
    end
  end

  # Find an in-progress transfer to this share
  def transferring_relay_session_for(share)
    @transferring_relay_sessions ||= {}
    @transferring_relay_sessions[share.id] ||=
      upload_requests_for_share(share).map do |request|
        request.e_sessions.detect{ |session| !session.ended? }  unless request.e_sessions.reverse.detect {|session| session.ended?}
      end.compact.first
  end

  # All upload requests for initial upload and relays
  def upload_requests_for_share(share)
    @upload_requests ||= {}
    @upload_requests[share.id] ||= TransferRequest.all(
      :conditions => [%q{
              e_package_id = ? AND
              share_id = ? AND
              direction IN (?, ?, ?)
              },
        e_package_id,
        share.id,
        ESession::DIRECTION_CREATE,
        ESession::DIRECTION_UPLOAD,
        ESession::DIRECTION_RELAY
      ],
      :order => 'created_at',
      :include => :e_sessions
    )
  end

  def upload_requests_since_last_connection(share)
    upload_requests = upload_requests_for_share(share)

    # Count from attempts since last successful connection (INCLUDING that attempt) -- error status = successful connection?
    last_connection = upload_requests.reverse.detect{ |r| r.e_sessions.any?{ |session| session.bytes_transferred > 0 } }
    upload_requests = upload_requests.select{ |request| request.created_at >= last_connection.created_at } if last_connection.present?

    upload_requests
  end

  def annealed_timeout_periods
    [
      2.minutes,
      2.minutes,
      5.minutes,
      30.minutes,
      1.hour,
      6.hours,
      6.hours
    ]
  end

  # Check for any completed upload
  def contents_exist_on_share?(share)
    upload_requests_for_share(share).any?{ |tr| tr.e_sessions.any?{ tr.status == ESession::STATUS_COMPLETE } }
  end

  def do_held_to_released_step_2
    #
    # We need to ensure all regular package routing and workgroup
    # relaying is done. (We don't worry about inter-faspex relays.)
    #
    # First we collect all of the destination shares. These come from
    # regular package routing *and* workgroup relays.
    #
    all_dest_shares = self.all_recipient_shares
    #
    # Next, we ensure a relay has succeeded for each share in the past
    # day or so.
    #
    # TODO: We should probably use a time stamp when the package is going live.
    # Here's the scenario: Package goes from held => live once. We then have
    # successful relay sessions stored in the database. Package goes
    # live => held. Meanwhile, old relay sessions are still in db. Package
    # goes held => ... almost live. We look at the database and see the
    # old relay sessions and think the routing is done, but it's not. So for
    # now, we just make sure the relay sessions aren't more than 1 day old. This
    # is not a permanent solution though.
    #
    count = 0
    errors = []
    all_dest_shares.each do |destination_share|
      #
      # We do this by ensuring every destination share has a completed
      # transfer request in the past day.
      #

      upload_requests = upload_requests_for_share(destination_share)

      if upload_requests.none?
        #TODO:  Put this back in and figure out what is going on
        #Rails.logger.error("Couldn't find relay transfer request for delivery with id #{self.id} and title #{self.title}, for share #{destination_share.name} ID:  #{destination_share.id.to_s}.")
        break
      end

      #
      # Check if it's done (there is one completed transfer session)
      #
      if upload_requests.any?{ |request| request.status == ESession::STATUS_COMPLETE }
        # Successful upload
        count += 1
      elsif upload_requests.all? { |request| [ESession::STATUS_ERROR, ESession::STATUS_TIMED_OUT, ESession::STATUS_VALIDATION_TIMED_OUT].include?(request.status) }
        # All upload attempts have failed - Retry or give up

        # Only consider uploads since last connection
        upload_requests = upload_requests_since_last_connection(destination_share)

        if upload_requests.length > annealed_timeout_periods.length
          # Tried several times - give up and move on
          errors << "RELAY: Gave up on relay to share #{destination_share.id} (#{destination_share.name}) for package #{self.id}"
          count += 1
        else
          last_session_stopped_at = upload_requests.map{ |request| request.e_sessions.map(&:e_session_stop_notification_time).compact.max || request.created_at }.max
          # Try again after a period of time
          if last_session_stopped_at < annealed_timeout_periods[upload_requests.length - 1].seconds.ago
            request = upload_requests.first
            Rails.logger.warn "RELAY: Attempting retry of relay to share #{destination_share.id} (#{destination_share.name}) for package #{self.id}"
            if request.flatten?
              # This was a relay
              e_workgroup = request.e_workgroup
              ewf = e_workgroup.e_workgroup_forwardings.find_by_share_id(destination_share.id)
              self.e_package.relay_to_forwarding_destination(destination_share, self, e_workgroup, ewf)
            else
              # This was an inbox upload - force unless there is a completed session to that share
              force = !contents_exist_on_share?(destination_share)
              self.relay_to_inbox(destination_share, force)
            end
          end
        end

      end
    end
    if count < all_dest_shares.length
      #
      # Regular package routing still not done yet. Reset transient state
      # back to 'routing packages'.
      #
      # TODO: Need to handle case when relaying to one or more shares fails.
      #
      self.update_state(
          Delivery::TRANSIENT_STATE_HELD_TO_RELEASED_ROUTING_PACKAGES)
      return
    end

    errors.each { |error| Rails.logger.error error }

    # Send received emails for any failed relays
    failed_relay_shares.each do |share|
      self.send_received_notifications(share)
    end

    #
    # --------------------------------------------------------------
    #
    # Package has been routed to all recipients and email should be
    # enqueued (the email background process fires off the emails).
    #
    # TODO: Deliver notification to creator that package is live?
    #
    # Mark as live
    #
    # Set the expiration date if packages expire
    if EConfiguration.packages_expire? && !self.expiration_date
      after_release = EConfiguration.get_days_until_packages_expire
      self.expiration_date = e_package.e_package_upload_completion_time + after_release.days
    end

    self.update_state(Delivery::STATE_RELEASED)
  end

  def self.get_deliveries_for_held_to_released_step_2
    deliveries = []

    Delivery.transaction do
      #
      # Look up all deliveries that are in the HELD_TO_RELEASED_ROUTING_PACKAGES
      # transient state
      #
      deliveries = Delivery.all(:conditions => [%q{
state = ?
},
                                                Delivery::TRANSIENT_STATE_HELD_TO_RELEASED_ROUTING_PACKAGES])
      #
      # Put them in transient state.
      #
      deliveries.each do |d|
        d.update_state(Delivery::TRANSIENT_STATE_HELD_TO_RELEASED_FINISH)
      end
    end

    deliveries
  end

  def self.handle_held_to_released_step_2

    deliveries = self.get_deliveries_for_held_to_released_step_2
    #
    # Process routing => live for each.
    #
    deliveries.each do |d|
      d.do_held_to_released_step_2
    end

  end

  # ----------------------------------------------------------------------
  # ----------------------------------------------------------------------

  ##
  # THIS IS EXECUTED BY THE 'RUN LOOP' IN
  #
  #            lib/daemons/background/background_worker.rb
  #
  # NOTE: We use DB transactions in a few places so that checking and
  # updating a package's state is atomic - where it matters.
  def self.handle_delivery_state_transitions
    #
    # We check for expired deliveries first.
    #
    # Held => Expired
    #
    self.handle_held_to_expired_deliveries
    #
    # Released => Expired
    #
    self.handle_released_to_expired_deliveries
    #
    # Held => Released, step 1: package routing
    #
    self.handle_held_to_released_step_1
    #
    # Held => Released, step 2: ensure all routing completed
    # and mark delivery as live.
    #
    self.handle_held_to_released_step_2

    # Deliveries whose package has been deleted
    self.handle_deleted_deliveries
  end

  # ----------------------------------------------------------------------
  # Miscellaneous
  # ----------------------------------------------------------------------

  ##
  # Combines all delivery and e_package error messages from attempted
  # database create/update.
  #
  # Taken from PackageMethods.
  def formatted_error_messages
    error_messages = self.errors.full_messages +
        self.e_package.errors.full_messages
    error_messages.join("\n")
  end

  ##
  # (Inner) directory where content is stored on the delivery's storage
  # share.
  #
  # This is just the e_package's inner directory.
  def storage_directory_path
    self.e_package.inner_directory_path
  end

  def held?
    self.state == Delivery::STATE_HELD
  end

  def live?
    self.state == Delivery::STATE_LIVE
  end

  ##
  # Checks if the number of successful downloads - of any kind, full or
  # partial - is at the download_limit.
  def download_limit_reached?
    #
    # If download_limit is null, number of downloads is unlimited.
    #
    if self.download_limit == nil
      return false
    end
    #
    # Locate all download sessions for the delivery whose
    # status is Complete.
    #
    download_count = ESession.count(:conditions => [
                                        'delivery_id = ? AND ' +
                                            'direction = ? AND ' +
                                            'status = ?',
                                        self.id, ESession::DIRECTION_DOWNLOAD,
                                        ESession::STATUS_COMPLETE
                                    ])
    #
    # Return
    #
    download_count >= self.download_limit
  end

  ##
  #
  # TODO: Update this doc
  #
  # Returns the remote_host a client should use for downloading the
  # Delivery's e_package (the remote_host for the e_package's
  # storage_share).
  #
  # TODO: Need to ensure this works properly for package routing.
  #
  # If the downloading user is creator of the package, the e_package's
  # storage_share should be the user's assigned share.
  #
  # **** But what if the downloading user forwarded the package?
  #
  # The assumption is that this method will be invoked when we are
  # generating a download transfer spec for the package sender. This occurs
  # along two paths:
  #    [1] When the user does #show on a package they sent (via /sent). The
  #        links inside the page refer to the /deliveries endpoint with
  #        the Delivery for the package. The /deliveries endpoint will
  #        use Delivery#remote_host to generate the transfer spec for
  #        the sender to download the package they sent.
  #    [2] The Faspex 2.0 package atom feed uses Delivery#remote_host to
  #        generate the transfer spec for the sender when they are listing
  #        their sent packages in their outbox.

  # Note that this may return the node's alternate address (if alternate
  # addresses are configured for the node). This is why the
  # downloading_client_ip_address and server_hostname args are necessary.
  #
  # This may throw an exception if the storage_share was deleted from
  # the database after the RecipientDelivery was created.
  def remote_host(downloading_client_ip_address, server_hostname)
    self.default_download_location.node.get_server_address(downloading_client_ip_address, server_hostname)
  end
end


