# Methods for setting up transfers go here, including generating fasp URLs,
# transfer tokens, and cookies

# including class must define e_package, delivery, and storage_share
module UrlGeneration
  # Options:
  # 'token' - whether or not to provide a token in the transfer spec, defaults to false
  # 'host' - alternate hostname to use in place of host returned by node
  # 'for_fasp_url' - whether this transfer spec is to be used for generating a fasp URL
  def get_transfer_spec(direction, paths_in_pkg, requester, options = {})
    # If direction is download, we need to see if we reached the download limit
    if direction == ESession::DIRECTION_DOWNLOAD
      if self.delivery.download_limit_reached?
        raise Delivery::TransferSpecError, 'Download limit reached'
      elsif self.delivery.release_policy == Delivery::RELEASE_POLICY_NEVER ||
            self.delivery.release_policy == Delivery::RELEASE_POLICY_ON_DATE &&
            self.delivery.release_date > Time.now &&
            self.delivery.sender != requester
        # Unreleased packages may only be downloaded by the sender
        raise Delivery::TransferSpecError,
          'Unable to download unreleased package' unless requester == self.delivery.sender
      # elsif self.delivery.e_package.e_package_upload_status != ESession::STATUS_COMPLETE
      #   raise Delivery::TransferSpecError, 'Unable to download package that has not finished uploading'
      end
    end

    # APPLE API
    if options[:share_id_order] and not options[:share_id_order].empty?
      default_destination_share = Share.default_destination
      target_share = ( (storage_share != default_destination_share) && !storage_share.safe_to_edit? ) ?     storage_share : default_destination_share
      spec = nil
      options[:share_id_order].each do |share_id|
        begin
          logger.warn("Package Routing: Checking share #{share_id.to_s} for presence of package #{e_package.id}")
          source_share = Share.find_by_id(share_id)
          #logger.warn("Share " + share_id.to_s + " is on node " + source_share.node.name)
          #logger.warn("Before path: " + e_package.node_path("/"))
          e_package.storage_share = source_share
          #logger.warn("After path: " + e_package.node_path("/"))
          transfer_settings_owner = target_share.override_transfer_parameters ? target_share : requester
          params = requested_params(direction, paths_in_pkg, requester, transfer_settings_owner, target_share, options[:for_fasp_url].present?)
          spec = source_share.node.get_transfer_spec(direction, params)

          path = spec.paths[0]['source']
          #logger.warn('Checking share for PATH: ' + path)
          response = source_share.node.browse(path,{})
          found = File.dirname(response.directory_items[0].path) + '/'
          #logger.warn('Found PATH: ' + found)
        rescue
          logger.warn("Package Routing Problem verifying package #{e_package.id} on share #{share_id.to_s}, skipping...")
        end
        next if nil == spec or nil == found
        options.except!(:host) if options[:host].present?
        spec.remote_host = source_share.node.get_server_address options[:request_ip] , options[:request_hostname]
        #logger.warn("HOST IS: " + options[:host])
        #logger.warn("Package Routing: Found package on share #{share_id.to_s}, generating spec")
        #logger.warn("Share info: #{source_share.to_yaml}")

        #add the source pairs back;
        spec.paths = get_source_paths_pairs(paths_in_pkg) if direction == ESession::DIRECTION_DOWNLOAD
        spec.paths = get_upload_source_paths_pairs(paths_in_pkg, options[:http_gateway_transfer].present?) if direction == ESession::DIRECTION_UPLOAD && spec.paths.blank?
        break
      end
    else
      # NORMAL CASE
      # Set target_share to share option only present in relays, otherwise set according to ESession direction
      if options[:share].present?
        target_share = Share.find(options[:share])
      elsif direction == ESession::DIRECTION_DOWNLOAD
        target_share = default_download_location
      else
        target_share = default_destination
      end
      transfer_settings_owner = target_share.override_transfer_parameters ? target_share : requester
      params = requested_params(direction, paths_in_pkg, requester, transfer_settings_owner, target_share, options[:for_fasp_url].present?, options[:http_gateway_transfer].present?)
      spec = target_share.node.get_transfer_spec(direction, params)

      #add the source pairs back;
      spec.paths = get_source_paths_pairs(paths_in_pkg, target_share) if direction == ESession::DIRECTION_DOWNLOAD
      spec.paths = get_upload_source_paths_pairs(paths_in_pkg, options[:http_gateway_transfer].present?) if direction == ESession::DIRECTION_UPLOAD && spec.paths.blank?
    end

    raise 'Unable to get transfer spec from storage node' if spec.nil?
    spec.remote_host = options[:host] if options[:host].present?
    spec.token = nil unless options[:token]
    spec
  end

  # Generate download path for all the source path in the package
  # @param paths_in_pkg selected paths to download
  # @returns array of source paths
  def get_source_paths_pairs(paths_in_pkg, target_share = nil)
    sources = paths_in_pkg.map do |path_in_pkg|
      { 'source' => download_path(path_in_pkg, target_share) }
    end

    return sources
  end

  def get_upload_source_paths_pairs(paths_in_pkg, http_gateway_transfer = false)
    obfuscate = http_gateway_transfer && require_obfuscation?
    pkg_path = node_path(paths_in_pkg.first)
    source_paths = e_package.e_package_paths.map(&:e_uploader_local_path)
    if http_gateway_transfer
      pairs = source_paths.map do |source_path|
      {
          'source' => source_path,
          'destination' => obfuscate ? obfuscate_filename(source_path).gsub(/=/, "") : source_path
      }
      end
    else
      pairs = source_paths.map do |source_path|
      {
        'source' => source_path
      }
      end
    end
    return pairs
  end

  def prepare_transfer(direction, paths_in_pkg, requesting_user, options = {} )
    transfer_cookie = TransferCookie.for_transfer(direction, requesting_user, delivery)
    TransferRequest.create(
      user: requesting_user,
      delivery: delivery,
      e_package: e_package,
      share: options.delete(:share) || storage_share,
      cookie: transfer_cookie.to_s,
      uuid: transfer_cookie.transfer_uuid,
      direction: direction,
      partial_package: direction == ESession::DIRECTION_DOWNLOAD ? !e_package.entire?(paths_in_pkg) : false,
      flatten: options.delete(:flatten) || false,
      e_workgroup_id: options.delete(:e_workgroup_id)
    )
  end

 private

  def select_transfer_rate_settings(direction, transfer_settings_owner, requester)
    transfer_rate_settings = { max_rate: requester.max_rate(direction), target_rate: requester.target_rate(direction), lock_minrate_policy: requester.lock_minrate_policy }
    if (requester != transfer_settings_owner) && transfer_settings_owner.override_transfer_parameters
      transfer_rate_settings = { target_rate:  [transfer_settings_owner.target_rate(direction), requester.target_rate(direction)].min,
                                 max_rate:  [transfer_settings_owner.max_rate(direction), requester.max_rate(direction)].min,
                                 lock_minrate_policy: (transfer_settings_owner.lock_minrate_policy | requester.lock_minrate_policy) }
    end
    transfer_rate_settings
  end

  def node_paths(paths_in_pkg)
    paths_in_pkg.map { |path| e_package.node_path(path) }
  end

  def create_dir?(direction, paths_in_pkg)
    direction != ESession::DIRECTION_DOWNLOAD || paths_in_pkg.count > 1
  end

  def protect?(direction)
    if direction == ESession::DIRECTION_UPLOAD
      require_ear? || e_package.ear_encrypted?
    else
      e_package.ear_encrypted?
    end
  end

  def content_protection(direction)
    case
    when direction == ESession::DIRECTION_UPLOAD && (require_ear? || e_package.ear_encrypted?)
      "encrypt"
    when direction == ESession::DIRECTION_DOWNLOAD && e_package.ear_encrypted?
      "decrypt"
    else
      nil
    end
  end

  def require_ear?
    if dropbox_delivery?
      dropbox.require_ear?
    elsif EConfiguration.can_override_ear_per_personal_invite? && !delivery.invitation.nil?
      delivery.ear_enabled
    else
      EConfiguration.require_ear?
    end
  end

  def require_obfuscation?
    EConfiguration.require_obfuscation? || delivery.obfuscation_enabled?
  end

  def obfuscate_filename(source)
      extension = source.slice(File.extname(source))
      obfuscated_file_name = ActiveSupport::Base64.urlsafe_encode64(source)
      obfuscated_file_name << extension
      obfuscated_file_name
  end


  def requested_params(direction, paths_in_pkg, requester, transfer_settings_owner, target_share, for_fasp_url = false, http_gateway_transfer = false)
    conf = CConfiguration.first
    node = target_share.node
    transfer_rate_settings = select_transfer_rate_settings(direction, transfer_settings_owner, requester)

    params = {
      "cookie" => prepare_transfer(direction, paths_in_pkg, requester).cookie,
      "target_rate_kbps" => transfer_rate_settings[:target_rate],
      "cipher_allowed" => conf.cipher,
      "content_protection" => content_protection(direction),
      "http_fallback" => conf.http_fallback_enabled? && !prevent_http?(direction),
      "create_dir" => create_dir?(direction, paths_in_pkg),
      "lock_rate_policy" => transfer_rate_settings[:lock_minrate_policy],
      "lock_min_rate" => transfer_rate_settings[:lock_minrate_policy],
      "target_rate_cap_kbps" => transfer_rate_settings[:max_rate],
      "tags" => tag_hash(direction == ESession::DIRECTION_DOWNLOAD, requester),
    }.merge!(path_and_root_params(direction, paths_in_pkg, for_fasp_url, http_gateway_transfer, target_share))

    #get token based on destination root and not source pairs
    if direction == ESession::DIRECTION_UPLOAD
      #use destination root instead of source and destination pairs
      params.merge!("destination_root" => node_path(paths_in_pkg.first))
    end

    # if transfer server is local, use our own http/https port instead of the one the
    # node daemon thinks is the right port
    if node.local? || FASPEX_CONFIG['UseApachePortsForHttpFallback']
      params["http_fallback_port"] = EConfiguration.first.e_http_port.to_i
      params["https_fallback_port"] = EConfiguration.first.e_https_port.to_i
    end

    TransferParams.new(params)
  end

  def prevent_http? direction
    if direction == ESession::DIRECTION_DOWNLOAD
      delivery.prevent_http_download
    elsif direction == ESession::DIRECTION_UPLOAD
      (delivery.invitation && delivery.invitation.prevent_http_upload)
    end
  end

  def path_and_root_params(direction, paths_in_pkg, exclude_sources, http_gateway_transfer, target_share = nil)
    if direction == ESession::DIRECTION_DOWNLOAD
      #do not send source pairs prior to the Node call to obtain transfer token
      # FASPX-2081 - fixes vulnerability that allowed file download from root directory using a package token
      first_path = download_path('/', target_share)
      {'paths' => [ "source" => first_path ]}
    elsif exclude_sources
      pkg_path = node_path(paths_in_pkg.first)

      pkg_path.gsub!(/^\//, '').split('/').inject('') do |path, directory|
        storage_share.api_client.create_directory(File.join(path, directory))
        path = File.join(path, directory)
      end

      if !http_gateway_transfer
        {'paths' => [{'destination' => pkg_path}]}
      else
      {'paths' => []}
      end
    else
      {'paths' => []}
    end
  end

  public

  def user_params(user)
    return {} if (user.nil? || user.is_a?(EWorkgroup))
    params={}

    if FASPEX_CONFIG['UserFieldsInTags'].nil?
      allowed = []
    else
      allowed=FASPEX_CONFIG['UserFieldsInTags'].split(/\s*,\s*/).map(&:downcase)
    end
    params[:name] = user.name if allowed.include?('name')
    params[:first_name] =  user.first_name if allowed.include?('first_name')
    params[:last_name] = user.last_name if allowed.include?('last_name')
    params[:email] =  user.email if allowed.include?('email')
    user.user_profile_data.each do |attribute|
      params[attribute.user_profile_field.name] = attribute.value if allowed.include?(attribute.user_profile_field.name.downcase)
    end

    params
  end


end
